from positron_common.common_dump import common_dump
from positron_common.observability.main import track_command_usage

# robbie "dump" command
@track_command_usage("dump")
def dump():
    common_dump()  # type: ignore





from pydantic import BaseModel
from .env_defaults import current
from .user_config import user_config
from typing import Optional
from positron_common.cli.console import console, print_boxed_messages

class EnvConfig(BaseModel):
    """
    The environment configuration for the Positron CLI.
    Loaded like so: defaults <- config file <- env vars.
    """
    API_BASE: str = current.api_base
    SOCKET_IO_DOMAIN: str = current.ws_base
    AUTH0_DOMAIN: str = current.auth0_domain
    AUTH0_CLIENT_ID: str = current.auth0_client_id
    AUTH0_AUDIENCE: str =  current.auth0_audience
    COMPRESSED_WS_NAME: str = 'workspace.tar.gz'
    SOCKET_IO_PATH: str = '/api/ws/socket.io'
    USER_AUTH_TOKEN: Optional[str] = None

    def __init__(self):
        super().__init__()

        # Override app dev defaults
        if user_config.user_auth_token:
            self.USER_AUTH_TOKEN = user_config.user_auth_token
        if user_config.backend_api_base_url:
            self.API_BASE = user_config.backend_api_base_url
        if user_config.backend_ws_base_url:
            self.SOCKET_IO_DOMAIN = user_config.backend_ws_base_url

    def to_string(self, include_title: bool = False) -> str:
        message = f"""- API_BASE: {self.API_BASE}
- SOCKET_IO_DOMAIN: {self.SOCKET_IO_DOMAIN}
- AUTH0_DOMAIN: {self.AUTH0_DOMAIN}
- AUTH0_CLIENT_ID: {self.AUTH0_CLIENT_ID}
- AUTH0_AUDIENCE: {self.AUTH0_AUDIENCE}
- COMPRESSED_WS_NAME: {self.COMPRESSED_WS_NAME}
- SOCKET_IO_PATH: {self.SOCKET_IO_PATH}
- USER_AUTH_TOKEN: {self.USER_AUTH_TOKEN}"""
        if include_title:
            return f"========== EnvConfig (env)==========\n{message}"
        else:
            return message
        
env = EnvConfig()
"""
The environment configuration for the Positron CLI.
"""



from positron_cli.login import login
from positron_common.job_api.validate_user_auth_token import is_auth_token_valid
from positron_common.user_config import user_config
from robbie.run_notebook import run_notebook

def init(loglevel: str = None):
    if not user_config.user_auth_token or not is_auth_token_valid():
        # there is no auth token set in .robbie/config.yaml or the token is invalid/expired in the backend
        login()
    run_notebook(loglevel)



from enum import Enum

# Cross component enums
log_types = dict(
    stdout="INFO",
    stderr="ERROR",
    debug="DEBUG"
)

class S3_CONSTANTS:
    FUNCTION_FOLDER = "function"
    SERIALIZED_FUNCTION_FILENAME = "function.pkl"
    SERIALIZED_ARGS_FILENAME = "args.pkl"
    SERIALIZED_KWARGS_FILENAME = "kwargs.pkl"
    RESULTS_FOLDER = "result"
    FUNCTION_RESULT_FILENAME = "result.pkl"
    FUNCTION_EXCEPTION_FILENAME = "exception.pkl"
    RESULTS_METADATA_FILENAME = "results_metadata.json"
    METADATA_FILENAME = "metadata.json"


class JobStatus(str, Enum):
    pending="pending"
    uploading="uploading"
    in_queue="in_queue"
    launching="launching_pod"
    pulling_image="pulling_image"
    pulled_image="pulled_image"
    starting_container="starting_container"
    started_container="started_container"
    initializing="initializing"
    computing="computing"
    storing="storing"
    complete="complete"
    failed="failed"
    execution_error="execution_error"
    terminate="terminate_job"
    terminated="terminated"

finished_job_statuses = {JobStatus.complete, JobStatus.execution_error, JobStatus.terminated, JobStatus.failed}
failed_job_statuses = {JobStatus.failed, JobStatus.execution_error, JobStatus.terminated}

class JobRunType(Enum):
    BASH_COMMAND_RUNNER = 'BASH_COMMAND_RUNNER'
    REMOTE_FUNCTION_CALL = 'REMOTE_FUNCTION_CALL'



"""A matplotlib backend for publishing figures via display_data"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the BSD 3-Clause License.
## Adapted fromt the matplotlib inline backend
## https://github.com/ipython/matplotlib-inline/blob/main/matplotlib_inline/backend_inline.py

import sys
import matplotlib
from matplotlib.backends import backend_agg
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib._pylab_helpers import Gcf
from matplotlib.figure import Figure

import datetime
import time

def get_filestring(extension='png'):
    # Get current time with nanoseconds
    now = datetime.datetime.now()
    nanoseconds = time.time_ns() % 1000000000

    # Format the datetime string
    datetime_str = now.strftime('%Y-%m-%d_%H:%M:%S.%f')
    datetime_str_with_ns = f"{datetime_str}{nanoseconds:09d}"

    return datetime_str_with_ns + '.' + extension


def new_figure_manager(num, *args, FigureClass=Figure, **kwargs):
    """
    Return a new figure manager for a new figure instance.

    This function is part of the API expected by Matplotlib backends.
    """
    return new_figure_manager_given_figure(num, FigureClass(*args, **kwargs))


def new_figure_manager_given_figure(num, figure):
    """
    Return a new figure manager for a given figure instance.

    This function is part of the API expected by Matplotlib backends.
    """
    manager = backend_agg.new_figure_manager_given_figure(num, figure)

    # Hack: matplotlib FigureManager objects in interacive backends (at least
    # in some of them) monkeypatch the figure object and add a .show() method
    # to it.  This applies the same monkeypatch in order to support user code
    # that might expect `.show()` to be part of the official API of figure
    # objects.  For further reference:
    # https://github.com/ipython/ipython/issues/1612
    # https://github.com/matplotlib/matplotlib/issues/835

    # Override show to save the figure as a file
    figure.show = lambda *a: figure.savefig(get_filestring())

    # ensure current figure will be drawn, and each subsequent call
    # of draw_if_interactive() moves the active figure to ensure it is
    # drawn last
    try:
        show._to_draw.remove(figure)
    except ValueError:
        # ensure it only appears in the draw list once
        pass
    # Queue up the figure for drawing in next show() call
    show._to_draw.append(figure)
    show._draw_called = True

    return manager


def show(close=None, block=None):
    """Show all figures as SVG/PNG payloads.

    Parameters
    ----------
    close : bool, optional
        If true, a ``plt.close('all')`` call is automatically issued after
        sending all the figures. If this is set, the figures will entirely
        removed from the internal list of figures.
    block : Not used.
        The `block` parameter is a Matplotlib experimental parameter.
        We accept it in the function signature for compatibility with other
        backends.
    """
    try:
        for figure_manager in Gcf.get_all_fig_managers():
            if figure_manager.canvas.figure:
                figure_manager.canvas.figure.savefig(get_filestring())

    finally:
        show._to_draw = []
        # only call close('all') if any to close
        # close triggers gc.collect, which can be slow
        if close and Gcf.get_all_fig_managers():
            matplotlib.pyplot.close('all')


# This flag will be reset by draw_if_interactive when called
show._draw_called = False
# list of figures to draw when flush_figures is called
show._to_draw = []


def flush_figures():
    """Send all figures that changed

    This is meant to be called automatically and will call show() if, during
    prior code execution, there had been any calls to draw_if_interactive.

    This function is meant to be used as a post_execute callback in IPython,
    so user-caused errors are handled with showtraceback() instead of being
    allowed to raise.  If this function is not called from within IPython,
    then these exceptions will raise.
    """
    if not show._draw_called:
        return

    try:
        # exclude any figures that were closed:
        active = set([fm.canvas.figure for fm in Gcf.get_all_fig_managers()])
        for fig in [fig for fig in show._to_draw if fig in active]:
            try:
                fig.savefig(get_filestring())
            except Exception as e:
                print("Exception in callback %r: %s" % (fig, e), file=sys.stderr)
    finally:
        # clear flags for next round
        show._to_draw = []
        show._draw_called = False


# Changes to matplotlib in version 1.2 requires a mpl backend to supply a default
# figurecanvas. This is set here to a Agg canvas
# See https://github.com/matplotlib/matplotlib/pull/1125
FigureCanvas = FigureCanvasAgg



import os
import threading
import select
import dill
import subprocess
import shutil
from IPython.core.magic import (Magics, magics_class, cell_magic)
from importlib.resources import files
from robbie.requirements_detection.main import show

LOCAL_PICKLE = 'cell.pkl'
REMOTE_PICKLE = 'cell_result.pkl'
REQUIREMENTS_FILENAME = 'cell_requirements.txt'
CELL_EXECUTABLE_FILENAME = 'cell_output.py'
JOB_RESULTS = 'job-execution'

do_not_load = ['open', 'os', 'do_not_load', 'load_pickle', 'local_pickle']

def local_pickle(user_ns: dict):
    ns_copy = user_ns.copy()
    baditems = dill.detect.baditems(user_ns)
    for baditem in baditems:
        ns_copy = {key: value for key, value in ns_copy.items() if value != baditem and key not in do_not_load}
    
    file = open(LOCAL_PICKLE, 'wb')
    protocol = dill.settings['protocol']
    pickler = dill.Pickler(file, protocol)
    pickler._byref = False   # disable pickling by name reference
    pickler._recurse = False # disable pickling recursion for globals
    pickler._session = True  # is best indicator of when pickling a session
    pickler._first_pass = True
    pickler.dump(ns_copy)
    file.flush()
    file.close()

def load_pickle(filename):
    file = open(filename, 'rb')
    unpickler = dill.Unpickler(file)
    user_ns = unpickler.load()
    file.close()
    return user_ns

# Registers this class name as a magic definition with ipython.
# Run with %%robbie
@magics_class
class robbie_magic(Magics):
    def write_requirements(self, filename: str):
        show(html=False, write_req_file=True, req_file_name=filename, caller_globals=self.shell.user_ns)
        
    
    @cell_magic
    def robbie(self, _, cell):
        # Ensure dill is installed and save the current session to a pickle file
        local_pickle(self.shell.user_ns)

        # get deps and ensure we'll install them.
        self.write_requirements(REQUIREMENTS_FILENAME)


        # create a file that wraps the cell and loads the pickle file
        exec_wrap_file = files('robbie').joinpath('executable_wrapper.py')
        exec_wrap_contents = exec_wrap_file.read_text().strip()
        exec_text = exec_wrap_contents.replace('### {cell}', cell)
        with open(CELL_EXECUTABLE_FILENAME, 'w') as f:
            f.write(exec_text)
        
        # Create a job on the fly and run the cell file in robbie
        command = f'robbie run --download "all" --tail --y "pip install -r {REQUIREMENTS_FILENAME} && pip install dill && python {CELL_EXECUTABLE_FILENAME}"',
        p = subprocess.Popen(
            command,
            # use a job_config.yaml, this was done to use NERC which has much faster startup times.
            # f'robbie run --tail --y',
            text=True,
            shell=True,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        # wait for the job to finish and print the output
        log_stop_event = threading.Event()
        stdout_thread = threading.Thread(target=logging_thread, args=(p.stdout, "stdout", log_stop_event)); stdout_thread.start()
        stderr_thread = threading.Thread(target=logging_thread, args=(p.stderr, "stderr", log_stop_event)); stderr_thread.start()
        p.wait()
        log_stop_event.set()
        stdout_thread.join()
        stderr_thread.join()

        # Job results should come back as a pkl file. We load the results back into the ipython environment
        try:
            copy_ns = load_pickle(f'./{REMOTE_PICKLE}')
            for key, value in copy_ns.items():
                self.shell.user_ns[key] = value
        except Exception as e:
            print(f'Error loading results: {e}')
            pass

        # Clean up env
        rm(CELL_EXECUTABLE_FILENAME)
        rm(LOCAL_PICKLE)
        rm(REMOTE_PICKLE)
        rm(REQUIREMENTS_FILENAME)
        rmdir(JOB_RESULTS)


def remove_line(multiline_str, match) -> str:
    lines = multiline_str.splitlines()  # Split into lines
    filtered_lines = [line for line in lines if match not in line]  # Filter lines
    return "\n".join(filtered_lines)  # Join the remaining lines

def rmdir(dirname):
    try:
        shutil.rmtree(dirname)
    except FileNotFoundError:
        pass

def rm(filename):
    try:
        os.remove(filename)
    except FileNotFoundError:
        pass
 
def logging_thread(pipe, stream_name: str, stop_event: threading.Event):
    try:
        with pipe:
            while not stop_event.is_set():
                # Calling pipe.readline() will hang the process until a new line is available
                # however, if no new lines are available, we want to check if the thread should stop
                ready, _, _ = select.select([pipe], [], [], 0.1)
                if ready:
                    # TODO: Can I iterate over the pipe and process everything that's ready?
                    line = pipe.readline()
                    if not line:
                        break
                    if line.strip():
                        if stream_name == "stdout":
                            print(line.rstrip())
                        else:
                            print("ERROR: " + line.rstrip())
                else:
                    stop_event.wait(1)

            # Read one last time.
            ready, _, _ = select.select([pipe], [], [], 0.1)
            if ready:
                for line in pipe:
                    if line.strip():
                        if stream_name == "stdout":
                            print(line.rstrip())
                        else:
                            print("ERROR: " + line.rstrip())
            # TODO: Do we need to check one more time for outstanding data in the pipe?
            # logger.debug(f'Logging thread for: {stream_name} stopped')
    except Exception as e:
        print(f'Logging thread for: {stream_name} stopped with exception: {e}')


# Loads the magic class into the runtime ipython environment.
# Load with %load_ext robbie.magic
def load_ipython_extension(ipython):
    """
    Any module file that define a function named `load_ipython_extension`
    can be loaded via `%load_ext module.path` or be configured to be
    autoloaded by IPython at startup time.
    """
    # You can register the class itself without instantiating it.  IPython will
    # call the default constructor on it.
    ipython.register_magics(robbie_magic)



from pydantic import BaseModel
from typing import Optional

class ErrorModel(BaseModel):
    code: Optional[str]

class ResponseErrorModel(BaseModel):
    userFriendlyErrorMessage: Optional[str]
    message: str
    error: Optional[ErrorModel]
    stack: Optional[str]



import zipfile
import tarfile
import os
from positron_common.compression import get_file_paths
from positron_common.constants import FILE_SIZE_THRESHOLD
from positron_job_runner.runner_env import runner_env
from positron_job_runner.cloud_storage import cloud_storage
from positron_job_runner.cloud_logger import logger

EXCLUDED_DIRECTORIES = ['venv', '.venv', '.git', '__pycache__', 'job-execution', '.robbie', '.ipynb_checkpoints', 'persistent-disk']
S3_BASE_PATH = f"{runner_env.JOB_OWNER_EMAIL}/{runner_env.JOB_ID}"
S3_RESULT_PATH = f"{S3_BASE_PATH}/result"

def unpack_workspace_from_s3():
    """Unpacks the workspace from S3"""
    s3_key = f"{S3_BASE_PATH}/workspace.tar.gz"
    if (runner_env.rerun and os.path.exists(runner_env.JOB_CWD)):
        rmdir_recursive(runner_env.JOB_CWD)
    try:
        local_tar_path = os.path.join(runner_env.RUNNER_CWD, 'workspace.tar.gz')
        cloud_storage.download_file(s3_key, local_tar_path)

        logger.info('Unpacking workspace')
        with tarfile.open(local_tar_path) as tar:
            tar.extractall(runner_env.JOB_CWD)
        logger.info('Workspace unpacked successfully')

    except Exception as e:
        logger.error(f"Failed to unpack workspace from S3: {e}")

def upload_results_to_s3():
    """Uploads the results to S3"""
    try:
        logger.info('Copying results to cloud storage...')

        results_dir = runner_env.JOB_CWD
        os.makedirs(results_dir, exist_ok=True)

        result_files = get_file_paths(path=results_dir, excluded_dirs=EXCLUDED_DIRECTORIES)

        # Create a tar.gz of the result directory
        results_zip_file_name = f"{results_dir}/result.zip"
        with zipfile.ZipFile(results_zip_file_name, "w", zipfile.ZIP_DEFLATED) as zipf:
            for file in result_files:
                logger.debug(f"Adding to zip: {file.name}")
                zipf.write(file.full_path, arcname=file.name)

        file_size = os.path.getsize(results_zip_file_name)
        if (file_size >= FILE_SIZE_THRESHOLD):
            size_in_mb = round(file_size / (1024 * 1024), 2)
            logger.warn(f"Results Archive Size: {size_in_mb} Mb. It might take a long time to upload it.")

        logger.debug(f"Uploading to cloud storage: {results_zip_file_name}")
        cloud_storage.upload_file(results_zip_file_name, f"{S3_RESULT_PATH}/result.zip")

        # Upload raw files to S3
        for file in result_files:
            logger.debug(f"Uploading to cloud storage: {file.name}")
            s3_key = f"{S3_RESULT_PATH}/{file.name}"
            cloud_storage.upload_file(file.full_path, s3_key)

        logger.info('Results uploaded to S3 successfully')

    except Exception as e:
        logger.error(f"Failed to upload results to S3: {e}")

def rmdir_recursive(path: str):
    """Recursively removes a directory"""
    for root, dirs, files in os.walk(path, topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))
        for name in dirs:
            os.rmdir(os.path.join(root, name))
    os.rmdir(path)


from typing import Any, Callable, Dict, Tuple, Union
from positron_common.aws.s3_presigned_handler import S3PresignedHandler
from positron_common.deployment.meta_data import MetaData, FunctionHashes
from positron_common.deployment.img_result import ImageResult
from positron_common.utils import is_notebook
from .serializer import Serializer
from ..enums import S3_CONSTANTS

class StoredFunction():
    func: Callable[[], str]
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    result: Union[Any, None] = None
    exception: Union[Exception, None] = None
    result_meta: Union[MetaData, None] = None
    s_func: bytes
    s_args: bytes
    s_kwargs: bytes
    s_result: bytes
    s_exception: bytes
    s_meta: MetaData
    job_id: str = ""
    base_dir: str

    def __init__(self, func = None, args = None, kwargs = None) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs


    def serialize_function(self):
        self.s_func = Serializer.serialize(self.func)
        self.s_args = Serializer.serialize(self.args)
        self.s_kwargs = Serializer.serialize(self.kwargs)


    def create_function_metadata(self, hmac_key: str):
        self.s_meta = MetaData(sha256_hash=FunctionHashes(
            func_hash=Serializer.compute_hash(self.s_func, hmac_key),
            args_hash=Serializer.compute_hash(self.s_args, hmac_key),
            kwargs_hash=Serializer.compute_hash(self.s_kwargs, hmac_key),
        ))


    def deserialize_function(self):
        self.func = Serializer.deserialize(self.s_func)
        self.args = Serializer.deserialize(self.s_args)
        self.kwargs = Serializer.deserialize(self.s_kwargs)


    def serialize_results(self):
        if self.result is not None:
            self.s_result = Serializer.serialize(self.result)
        if self.exception is not None:
            self.s_exception = Serializer.serialize(self.exception)


    def create_results_metadata(self, hmac_key: str):
        sha256_hash=FunctionHashes()
        if self.result is not None:
            sha256_hash.result_hash = Serializer.compute_hash(self.s_result, hmac_key)
        if self.exception is not None:
            sha256_hash.exception_hash = Serializer.compute_hash(self.s_exception, hmac_key)

        self.s_meta = MetaData(sha256_hash=sha256_hash)


    def set_job_id(self, job_id: str):
        self.job_id = job_id


    def upload_to_s3(self):
        func_path = f"{S3_CONSTANTS.FUNCTION_FOLDER}/{S3_CONSTANTS.SERIALIZED_FUNCTION_FILENAME}"
        args_path = f"{S3_CONSTANTS.FUNCTION_FOLDER}/{S3_CONSTANTS.SERIALIZED_ARGS_FILENAME}"
        kwargs_path = f"{S3_CONSTANTS.FUNCTION_FOLDER}/{S3_CONSTANTS.SERIALIZED_KWARGS_FILENAME}"
        meta_path = f"{S3_CONSTANTS.FUNCTION_FOLDER}/{S3_CONSTANTS.METADATA_FILENAME}"
        S3PresignedHandler.upload_bytes_to_job_folder(self.s_func, self.job_id, func_path)
        S3PresignedHandler.upload_bytes_to_job_folder(self.s_args, self.job_id, args_path)
        S3PresignedHandler.upload_bytes_to_job_folder(self.s_kwargs, self.job_id, kwargs_path)
        S3PresignedHandler.upload_bytes_to_job_folder(self.s_meta.to_json(), self.job_id, meta_path)


    def run(self):
        """Runs the function and stores the result"""
        try:
            self.result = self.func(*self.args, **self.kwargs)
        except Exception as e:
            self.exception = e


    def load_and_validate_results(self, hmac_key: str):
        image_result = ImageResult(self.job_id)
        if is_notebook() and image_result.has_images():
            image_result.show_images()

        try:
            function_result_path = f"{S3_CONSTANTS.RESULTS_FOLDER}/{S3_CONSTANTS.FUNCTION_RESULT_FILENAME}"
            function_exception_path = f"{S3_CONSTANTS.RESULTS_FOLDER}/{S3_CONSTANTS.FUNCTION_EXCEPTION_FILENAME}"
            result_metadata_path = f"{S3_CONSTANTS.RESULTS_FOLDER}/{S3_CONSTANTS.RESULTS_METADATA_FILENAME}"
            self.s_result = S3PresignedHandler.download_file_to_bytes(self.job_id, function_result_path)
            self.s_exception = S3PresignedHandler.download_file_to_bytes(self.job_id, function_exception_path)
            result_meta = S3PresignedHandler.download_file_to_bytes(self.job_id, result_metadata_path)
            self.result_meta = MetaData.from_json(result_meta)
            self._validate_result_meta(hmac_key)
            self.result = Serializer.deserialize(self.s_result) if self.s_result is not None else None
            self.exception = Serializer.deserialize(self.s_exception) if self.s_exception is not None else None
        except Exception as e:
            self.exception = e
        return self.result, self.exception

    def _validate_result_meta(self, hmac_key: str):
        result_computed_hash = Serializer.compute_hash(self.s_result, hmac_key)
        exception_computed_hash = Serializer.compute_hash(self.s_exception, hmac_key)
        if (self.s_result):
            if self.result_meta.sha256_hash.get('result_hash') != result_computed_hash:
                raise Exception('Integrity check of result files failed. Ensure no one has access to your cloud storage.')
        if (self.s_exception):
            if self.result_meta.sha256_hash.get('exception_hash') != exception_computed_hash:
                raise Exception('Integrity check of result files failed. Ensure no one has access to your cloud storage.')




from functools import wraps
import sentry_sdk
from positron_common.build_env import build_env

def track_command_usage(command_name):
    """Decorator to track usage of commands in Sentry with arguments."""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Capture command usage event with arguments
            sentry_sdk.capture_event({
                "message": f"Command Run: {command_name}",
                "level": "info",
                "tags": {"command": command_name},
                "extra": {
                    "arguments": args,
                    "keyword_arguments": kwargs
                }
            })
            # Execute the command function
            return func(*args, **kwargs)
        return wrapper
    return decorator

def setup():
  """Should be run as early in the app lifecycle as possible"""
  # TODO: turn off in local dev mode.
  sentry_sdk.init(
    dsn="https://76683e3eb4abf2e208a27d252ab023bb@o4508017904910336.ingest.us.sentry.io/4508017906089984",
    # Set traces_sample_rate to 1.0 to capture 100%
    # of transactions for tracing.
    traces_sample_rate=1.0,
    # Set profiles_sample_rate to 1.0 to profile 100%
    # of sampled transactions.
    # We recommend adjusting this value in production.
    profiles_sample_rate=1.0,
    environment=build_env.value
  )

setup()


import requests
import json
from positron_common.exceptions import RemoteCallException
from ..env_config import env
from ..cli.logging_config import logger

def start_job(job_id, data):
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "PositronJobId": job_id}
    url = f'{env.API_BASE}/start-job'
    
    logger.debug(f'Calling: {url}')
    response = requests.post(url, headers=Headers, json=data)
    
    logger.debug(response)
    if response.status_code != 200:
        body = response.json()
        logger.debug(json.dumps(body, indent=2))
        if body.get('userFriendlyErrorMessage'):
            raise RemoteCallException(body.get('userFriendlyErrorMessage'))
        else:
            raise RemoteCallException(f'start job failed with http code: {response.status_code} \n {response.text}')
    else:
        logger.debug(json.dumps(response.json(), indent=2))
        return response.json()


import time
from .config import PositronJob
from .cli.console import console, ROBBIE_ORANGE, ROBBIE_BLUE
from rich.text import Text
from rich.panel import Panel
from rich import box
from .exceptions import RemoteCallException
from .cli.logging_config import logger

def print_robbie_configuration_banner(job: dict, job_config: PositronJob):
    text = Text()

    if job_config.name:
        text.append("Run Name: ", style=ROBBIE_BLUE)
        text.append(f"{job_config.name}\n")

    text.append("Python Version: ", style=ROBBIE_BLUE)
    text.append(f"{job_config.python_version}\n")

    text.append("Robbie Python SDK Version: ", style=ROBBIE_BLUE)
    text.append(f"{job_config.robbie_sdk_version}\n")

    text.append("Dependency file: ", style=ROBBIE_BLUE)
    text.append(f"{job_config.dependencies}\n")

    text.append("Conda Environment: ", style=ROBBIE_BLUE)
    text.append(f"{job_config.conda_env}\n")
    
    text.append("Funding Source: ", style=ROBBIE_BLUE)
    if not job_config.funding_selection:
        job_config.funding_selection = "Default"
    text.append(f"{job['fundingGroupName']} ({job['fundingGroupId']}) - {job_config.funding_selection}\n")
        
    text.append("Hardware: ", style=ROBBIE_BLUE)
    if not job_config.environment_selection:
        job_config.environment_selection = "Default"
    text.append(f"{job['environmentName']} ({job['environmentId']}) - {job_config.environment_selection}\n")

    if job_config.cluster:
        text.append("Cluster: ", style=ROBBIE_BLUE)
        text.append(f"{job_config.cluster}\n")
        
    text.append("Image: ", style=ROBBIE_BLUE)
    if not job_config.image_selection:
        job_config.image_selection = "Default" 
    text.append(f"{job['imageName']} - {job_config.image_selection}\n")
        
    text.append("Workspace Directory: ", style=ROBBIE_BLUE)
    text.append(f"{job_config.workspace_dir}\n") 
    text.append("Max Token Consumption: ", style=ROBBIE_BLUE)
    if not job["maxUsableTokens"]:
        text.append("Not specified\n")
    else:
        text.append(f"{job['maxUsableTokens']}\n")

    text.append("Max Execution Time (minutes): ", style=ROBBIE_BLUE)

    if not job["maxExecutionMinutes"]:
        text.append("Not specified\n")
    else:
        text.append(f"{job['maxExecutionMinutes']}\n")
    
    text.append("Environment Variables: ", style=ROBBIE_BLUE)

    if not job_config.env:
         text.append("None specified\n")
    else:
        text.append(f"{job_config.env}\n")


    if job_config.commands:
        text.append("Shell Commands:  \n", style=ROBBIE_BLUE)
        for cmd in job_config.commands:
            text.append(f'{cmd}\n')
    
    console.print(Panel(
        text,
        box=box.ROUNDED,
        # padding=(1, 2),
        title = Text(f"Robbie Run Configuration ({job['tokenRatePerHour']} tokens/hour)", style=ROBBIE_ORANGE),
        border_style=ROBBIE_ORANGE,
    ))
    logger.debug(f"========== Robbie Run Configuration ({job['tokenRatePerHour']} tokens/hour) ========== \n{text}")

def print_job_details_banner(job: dict):
    ## print job details
    text = Text()
    text.append("Run Name: ", style=ROBBIE_BLUE)
    text.append(f"{job['name']}\n")
            
    text.append("Run ID: ", style=ROBBIE_BLUE)
    text.append(f"{job['id']}\n")
        
    text.append("Start Time: ", style=ROBBIE_BLUE)
    text.append(f"{time.asctime()}")

    console.print(Panel(
        text,
        box=box.ROUNDED,
        title=Text("Run Details", style=ROBBIE_ORANGE),
        border_style=ROBBIE_ORANGE,
    ))
    logger.debug(f"========== Run Details ========== \n{text}")
    
# prints a rich job completion banner
def print_job_complete_banner(job: dict, start):
    ## print job details
    text = Text()
    text.append("Job Name: ", style=ROBBIE_BLUE)
    text.append(f"{job['name']}\n")
            
    text.append("Total time: ", style=ROBBIE_BLUE)
    text.append(f"{time.perf_counter() - start:.2f} seconds.\n")
        
    text.append("Tokens consumed: ", style=ROBBIE_BLUE)
    text.append(f"{job['tokensUsed']}\n")
        
    text.append("RESULT: ")
    if(job['status'] == "complete"):
        text.append(f"Success", style="green")
    else:
        text.append(f"{job['status']}", style="red")
                
    console.print(Panel(
        text,
        box=box.ROUNDED,
        # padding=(1, 2),
        title=Text("Run Complete", style=ROBBIE_ORANGE),
        border_style=ROBBIE_ORANGE,
    ))
    logger.debug(f"========== Run Complete ========== \n{text}")    

def print_known_error(e: RemoteCallException):
    logger.debug(e, exc_info=True)
    console.print(f"[red]An error has occurred: {e.user_friendly_message}[/red]")
    if e.additional_help:
        console.print(f"[yellow]{e.additional_help}[/yellow]")



import requests
from ..env_config import env
from ..cli.logging_config import logger

def is_auth_token_valid() -> bool:
    Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN}
    url = f'{env.API_BASE}/validate-user-auth-token'
    
    logger.debug(f'Calling: {url}')
    response = requests.get(url, headers=Headers)
    
    logger.debug(response)
    if response.status_code == 200:
        return True
    else:
        return False



import requests

from positron_common.env_config import env
from positron_common.exceptions import RemoteCallException, DecoratorException
from positron_common.constants import temp_path
from ..cli.logging_config import logger


class S3PresignedHandler():
    """This class is responsible for handling file transfer 
       to and from Amazon S3 trough presigned URLs."""

    @staticmethod
    def upload_bytes_to_job_folder(bytes, job_id, path):
        presigned_resp = S3PresignedHandler._get_upload_presigned_url(job_id, path)
        
        # we need to create a temporary file from the bytes to upload it to s3
        filename = path.split('/')[-1]

        with open(f"{temp_path}/{filename}", 'wb') as f:
            f.write(bytes)
        with open(f"{temp_path}/{filename}", 'rb') as f:
            logger.debug(f'Uploading {filename} to {presigned_resp.get("url")}')
            response = requests.post(presigned_resp.get('url'), data=presigned_resp.get('fields'), files={"file": (filename, f)}) 
            if response.status_code != 204:
                raise DecoratorException(
                    f'Upload failed with http code: {response.status_code} \n {response.text}')
            


    @staticmethod
    def upload_file_to_job_folder(file_path, job_id, path):
        presigned_resp = S3PresignedHandler._get_upload_presigned_url(job_id, path)

        with open(file_path, 'rb') as f:
            files = {'file': (path, f)}
            response = requests.post(presigned_resp.get('url'), data=presigned_resp.get('fields'), files=files)
            logger.debug(response)
            if response.status_code != 204:
                raise DecoratorException(
                    f'Upload failed with http code: {response.status_code} \n {response.text}')


    @staticmethod
    def download_file_to_bytes(job_id, s3_path):
        presigned_resp = S3PresignedHandler._get_download_presigned_url(job_id, s3_path)
        resp = requests.get(presigned_resp.get('url'))
        if resp.status_code != 200:
            return None
        else:
            return resp.content


    @staticmethod
    def _get_upload_presigned_url(job_id, filename):
        Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "PositronJobId": job_id}
        url = f'{env.API_BASE}/generate-presigned-url?filename={filename}'
        logger.debug(f'Calling: {url}')
        response = requests.get(url, headers=Headers)
        logger.debug(response)
        if response.status_code != 200:
            raise RemoteCallException(f'Presigned url fetching failed with http code: {response.status_code} \n {response.text}')
        else:
            logger.debug(response.json())
            return response.json()

    @staticmethod
    def _get_download_presigned_url(job_id, filename):
        Headers = {"PositronAuthToken": env.USER_AUTH_TOKEN, "PositronJobId": job_id}
        url = f'{env.API_BASE}/generate-download-url?filename={filename}'
        logger.debug(f'Calling: {url}')
        response = requests.get(url, headers=Headers)
        logger.debug(response)
        if response.status_code != 200:
            raise RemoteCallException(
                f'Presigned url fetching failed with http code: {response.status_code} \n {response.text}')
        else:
            logger.debug(response.json())
            return response.json()


import json
import logging
import time
import os
from logging import _nameToLevel
from typing import Optional
from .console import console

APP_NAME = "robbie.cli"
LOG_LEVEL = os.getenv("ROBBIE_LOG_LEVEL", "ERROR").upper()
JSON_LOGS=False
JSON_LOGS_PRETTY=False
JSON_LOGS_INCLUDE_SRC_LINE=False

log_level = _nameToLevel.get(LOG_LEVEL, 20)

class _UTCFormatter(logging.Formatter):
    """Class that overrides the default local time provider to GMT in log formatter."""
    converter = time.gmtime


    def format(self, record):
        if not JSON_LOGS:
            return super(_UTCFormatter, self).format(record)

        log_record = {
            "asctime": time.strftime("%Y-%m-%d %H:%M:%S:%MS", self.converter(record.created)),
            "name": record.name,
            "level": record.levelname,
            "message": record.getMessage(),
        }

        if JSON_LOGS_INCLUDE_SRC_LINE:
            relative_path = os.path.relpath(record.pathname)
            # Append the line number to the relative path
            relative_path_with_lineno = f"{relative_path}:{record.lineno}"
            log_record["pathname"] = relative_path_with_lineno

        if JSON_LOGS_PRETTY:
            return json.dumps(log_record, indent=4)
        return json.dumps(log_record)

logger = logging.getLogger(APP_NAME)

logger.setLevel(log_level)
handler = logging.StreamHandler()
formatter = _UTCFormatter("%(asctime)s %(name)s %(levelname)-8s %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

def set_log_level(loglevel: Optional[str]):
    if loglevel is None:
        return
    level_int = _nameToLevel.get(loglevel)
    if level_int is None:
        console.print("[red]Invalid log level")
        return
    logger.setLevel(level_int)
    console.print("[green]loglevel set to: ", loglevel)

# TODO: Not sure if we really need this.
# don't stream logs with the root logger handler
logger.propagate = 0

if __name__ == "__main__":
    """Testing"""
    logger.debug("Debug message")
    logger.info("Info message")
    logger.warning("Warning message")
    logger.error("Error message")
    logger.critical("Critical message")


import time
from dataclasses import dataclass
import os
import tarfile
from typing import List
from rich.tree import Tree
from rich.panel import Panel
from rich import box
from rich.text import Text
from positron_common.env_config import env
from positron_common.constants import temp_path
from rich.prompt import Confirm
from positron_common.cli.console import console, ROBBIE_YELLOW
from positron_common.cli.logging_config import logger
from positron_common.constants import FILE_SIZE_THRESHOLD

@dataclass
class FileEntry:
    full_path: str
    name: str
    size: int

def get_file_paths(path: str, excluded_dirs: List[str], excluded_files: List[str] = []) -> List[FileEntry]:
    """should return a list relative paths that should included in the results"""
    all_files: List[FileEntry] = []
    for root, dirs, files in os.walk(path):
        for exclude_dir in excluded_dirs:
            try:
                dirs.remove(exclude_dir)
            except ValueError:
                pass
        for file in files:
            if file in excluded_files:
                continue
            full_path = os.path.join(root, file)
            file_size = os.path.getsize(full_path)
            rel_path = os.path.relpath(full_path, path)
            all_files.append(FileEntry(name=rel_path, size=file_size, full_path=full_path))
    return all_files

def check_workspace_dir(workspace_dir: str):
    if not os.path.exists(workspace_dir):
        question_prompt = f"Directory '{workspace_dir}' does not exist. Would you like to create it?"
        if Confirm.ask(f'{question_prompt}', default=True):
            os.makedirs(workspace_dir)
            return True
        console.print("[yellow]Either update the 'workspace_dir' config or remove it to use the default of the current workspace.[/yellow]")
        return False
    return True

def create_workspace_tar(workspace_dir: str) -> int:
    tree = Tree("Local (workspace) files to copy to Remote Machine")
    path = f"{temp_path}/{env.COMPRESSED_WS_NAME}"
    last_files_compressed_count = 0

    if not os.path.exists(temp_path):
        os.makedirs(temp_path)

    excluded_dirs = ['venv', '.venv', '.git', '__pycache__', 'job-execution', ".robbie", ".ipynb_checkpoints", temp_path]
    files = get_file_paths(path=workspace_dir, excluded_dirs=excluded_dirs)

    logger.debug(f"Files to compress: {files}")
    with tarfile.open(path, 'w:gz') as tar:
        for file in files:
            if not file.name.endswith((".py", ".ipynb", ".yaml", ".csv", ".txt", ".pkl")):
                continue
            tree.add(f"[yellow]{file.name}, size: {file.size} bytes[/yellow]")
            last_files_compressed_count += 1
            tar.add(file.full_path, arcname=file.name)

    file_size = os.path.getsize(path)
    if (file_size >= FILE_SIZE_THRESHOLD):
        size_in_mb = round(file_size / (1024 * 1024), 2)
        text = Text(f"Size: {size_in_mb} Mb. It might take a long time to upload it.", style=ROBBIE_YELLOW)
        console.print(Panel(
            renderable=text,
            box=box.ROUNDED,
            title = Text("Workspace Archive Size Warning", style=ROBBIE_YELLOW),
            border_style=ROBBIE_YELLOW,
            padding=(1, 2),
        ))
        time.sleep(4)

    if last_files_compressed_count > 0:
        console.print(tree)
    return last_files_compressed_count



import requests
import time
import typer
from positron_common.env_config import env
from positron_common.cli.console import console
from positron_common.exceptions import RemoteCallException, RobbieException

def get_device_code_payload():
    device_code_payload = {
        'client_id': env.AUTH0_CLIENT_ID,
        'scope': 'openid profile',
        'audience': env.AUTH0_AUDIENCE
    }
    device_code_response = requests.post(f'https://{env.AUTH0_DOMAIN}/oauth/device/code', data=device_code_payload)
    if device_code_response.status_code != 200:
        raise RobbieException(f'Error requesting device code: {device_code_response.text}')

    return device_code_response.json()

def get_user_auth_token(access_token: str) -> None:
    auth_header = {
        'Authorization': f'Bearer {access_token}'
    }
    user_token_response = requests.get(f'{env.API_BASE}/get-user-auth-token', headers=auth_header)
    if user_token_response.status_code != 200:
        raise RemoteCallException.from_request(user_token_response)

    return user_token_response.json()

def wait_for_access_token(device_code_data):
    token_payload = {
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
        'device_code': device_code_data['device_code'],
        'client_id': env.AUTH0_CLIENT_ID
    }
    authenticated = False
    while not authenticated:
        console.print('Waiting for the user to login...')
        token_response = requests.post(f'https://{env.AUTH0_DOMAIN}/oauth/token', data=token_payload)

        token_data = token_response.json()
        if token_response.status_code == 200:
            console.print('Authenticated!')
            authenticated = True
            return token_data['access_token']
        elif token_data['error'] not in ('authorization_pending', 'slow_down'):
            print(token_data['error_description'])
            raise typer.Exit(code=1)
        else:
            time.sleep(device_code_data['interval'])



import select
import subprocess
import requests
import time
import threading
from datetime import datetime, timezone
from positron_common.exceptions import RobbieException
from positron_common.enums import JobRunType
from positron_job_runner.runner_env import runner_env
from positron_job_runner.cloud_logger import logger
from positron_job_runner.remote_function import RemoteFunction
from positron_job_runner.workspace import unpack_workspace_from_s3, upload_results_to_s3

'''_______________________________________________________________________________________
    
    INITIALIZING CONFIGURATION AND CONSTANTS
    ______________________________________________________________________________________
'''

# Define API endpoints
API_BASE = runner_env.API_ENDPOINT
API_GET_JOB = f'{API_BASE}/get-job'
API_UPDATE_JOB = f'{API_BASE}/update-job'
API_JOB_LIFECYCLE = f'{API_BASE}/check-lifecycle'

# Define cross component enums
job_statuses = dict(
    pending="pending",
    uploading="uploading",
    in_queue="in_queue",
    launching="launching_pod",
    pulling_image="pulling_image",
    pulled_image="pulled_image",
    starting_container="starting_container",
    started_container="started_container",
    initializing="initializing",
    computing="computing",
    storing="storing",
    complete="complete",
    failed="failed",
    execution_error="execution_error",
    terminate="terminate_job",
    terminated="terminated"
)

# Job management
running_job = None
final_status = job_statuses['complete']

# Request header
headers = {
    "PositronJobId": runner_env.JOB_ID,
    "SystemAuthenticationKey": runner_env.SYSTEM_AUTHENTICATION_KEY,
}

'''_______________________________________________________________________________________
    
    SUB MODULES
    ______________________________________________________________________________________
'''
# Get Job -----------------------------------------------------------
def get_job():
    logger.info('Fetching job details')
    resp = try_request(requests.get, API_GET_JOB, headers=headers)
    job = resp.json()
    return job

# Run Job -----------------------------------------------------------
def run_job(job: dict):
    logger.info('Starting job')
    update_job(status=job_statuses["computing"])
    
    job_type_str = job.get('entryPoint')
    logger.debug(f"Running job with type: {job_type_str}")
    job_type = JobRunType[job_type_str]
    if job_type == JobRunType.BASH_COMMAND_RUNNER:
        run_commands(job)
        return
    if job_type == JobRunType.REMOTE_FUNCTION_CALL:
        run_remote_function_job(job)
        return
    if job_type == None:
        raise RobbieException(f"Job type was not set. This could be a compatibility issue with the robbie library.")
    raise RobbieException(f"Unknown job type of: '{job_type}'!")


def run_commands(job: dict):
    command: str = job.get('commands')
    logger.debug(f'Found commands to execute:\n{command}')
    command_prefix = 'cd .. && python -m venv venv && . venv/bin/activate && cd job-execution'
    full_command = f"{command_prefix} && {command}"
    logger.debug(f"Full command: {full_command}")
    start_and_monitor_job(full_command)


def run_remote_function_job(job: dict):
    # Determine which cli to install
    robbie_local_cli_cmd = f"[ -d \"{runner_env.ROBBIE_CLI_PATH}\" ] && echo 'Installing local robbie cli...' && pip install {runner_env.ROBBIE_CLI_PATH}"
    robbie_env_cli_cmd = "echo 'Installing robbie' && pip install robbie"
    if (runner_env.POSITRON_CLI_ENV == "development"):
        robbie_env_cli_cmd = "echo 'Installing robbie:dev' && pip install --index-url https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple/ --pre robbie"
    if (runner_env.POSITRON_CLI_ENV == "test"):
        robbie_env_cli_cmd = "echo 'Installing robbie:test' && pip install --index-url https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple/ robbie"
    install_robbie_cmd = f"{robbie_local_cli_cmd} || {robbie_env_cli_cmd}"

    enter_env = "cd .. && python -m venv venv && . venv/bin/activate && cd job-execution"
    install_requirements = "[ -f requirements.txt ] && pip install -r requirements.txt || echo 'No requirements to install'"

    # Construct the command
    # By putting the install_robbie_cmd at the end, we ensure the latest version of robbie is installed,
    # overriding any version in the requirements.txt
    command_base = " && ".join([
        enter_env,
        install_requirements,
        install_robbie_cmd,
    ])

    logger.info('Downloading function and arguments')
    rm = RemoteFunction()
    rm.setup()

    meta = job.get('meta', {})
    job_arguments = meta.get('jobArguments', [])

    # Join job arguments with spaces
    arguments_string = ' '.join(job_arguments)

    # Construct the execution command
    execution_command = f"{rm.run_cmd} {arguments_string}"

    # Log the execution command
    logger.info(f"Installing Dependencies and running: {execution_command}")

    # Combine the base command with the execution command
    full_command = f"{command_base} && {execution_command}"
    logger.debug(f"Full command: {full_command}")
    start_and_monitor_job(full_command)


def start_and_monitor_job(command: str):
    sub_env = runner_env.env_without_runner_env()
    logger.debug(f'Running client job with job environment: {sub_env}')

    # Start job
    global running_job
    if (runner_env.JOB_USER != 'job_user'):
        logger.debug('Running job as current user')
        running_job = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,                # allows running on any os
            universal_newlines=True,   # otherwise stdout and stderr will be bytes
            cwd=runner_env.JOB_CWD,
        )
    else:
        logger.debug('Running job as protected user')
        running_job = subprocess.Popen(
            ["su", "-c", command, "-s", "/bin/bash", runner_env.JOB_USER],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            # requires elevated permissions which is not allowed in the runtime environment.
            # shell=True,
            cwd=runner_env.JOB_CWD,
            env=sub_env,
        )

    logger.debug(f"Job started with PID: {running_job.pid}")

    # Start parallel threads
    stop_event = threading.Event()
    log_stop_event = threading.Event()
    logger.debug("Starting threads...")
    termination_thread = start_termination_thread(stop_event)
    charge_thread = start_charging_thread(stop_event)
    stdout_thread = threading.Thread(target=logging_thread, args=(running_job.stdout, "stdout", log_stop_event)); stdout_thread.start()
    stderr_thread = threading.Thread(target=logging_thread, args=(running_job.stderr, "stderr", log_stop_event)); stderr_thread.start()

    logger.info("Waiting for job to complete")
    running_job.wait()
    return_code = running_job.returncode
    stop_event.set()
    if return_code == -9:
        # Force stop
        log_stop_event.set()
        logger.warn("Job process terminated. Shutting down open threads...")
        logger.warn("Some of your logs will be lost.")
    else:
        logger.info(f"Job completed with code: {return_code}. Waiting for threads to finish processing...")
        # Note: Streaming stdout will take much longer to process than the actual process itself.
        stdout_thread.join()
        stderr_thread.join()

    # Wait for charging and termination thread to finish.
    charge_thread.join()
    termination_thread.join()
    logger.debug("Job run completed.")

    global final_status
    return_code = running_job.returncode
    if final_status != job_statuses['terminated'] and return_code > 0: final_status = job_statuses["execution_error"]

def decode_if_bytes(line):
    return line.decode('utf-8') if isinstance(line, bytes) else line

# Finish Job --------------------------------------------------------
def finish_job(job: dict):
    logger.info("Finishing job")
    out_str = '\n'.join(decode_if_bytes(line) for line in logger.stdout)
    err_str = '\n'.join(decode_if_bytes(line) for line in logger.stderr)

    tokens_used = job['tokensUsed']

    if tokens_used:
        tokens_used_msg = f'Total tokens used for processing job: {tokens_used}'
        logger.info(tokens_used_msg)
        out_str += f'\n{tokens_used_msg}'

    logger.debug(f"Updating job with logs: output log size: {len(out_str)}, error log size: {len(err_str)}")

    update_job(status=final_status, end_date=datetime.now(timezone.utc), output_log=out_str, error_log=err_str)
    
    # Shared backend checks for this signal to terminate socket connection
    logger.info("Job successfully completed.")
    logger.end_of_logs_signal()


'''_______________________________________________________________________________________
    
    UTILS
    ______________________________________________________________________________________
'''
# Check for termination ---------------------------------------------
def is_job_active():
    if (runner_env.rerun):
        return True

    resp = try_request(requests.get, API_GET_JOB, headers=headers)
    db_job = resp.json()
    inactive_statuses = [
        job_statuses["complete"],
        job_statuses["failed"],
        job_statuses["execution_error"],
        job_statuses["terminate"],
        job_statuses["terminated"],
    ]

    if db_job['status'] not in inactive_statuses:
        return True

    global final_status
    if db_job['status'] == job_statuses["terminate"]:
        final_status = job_statuses["terminated"]
        logger.info("The job has been terminated by the user.")
    else:
        final_status = db_job['status']
        logger.info(f"The job is not active anymore (status: {db_job['status']})")

    return False


# Update Job --------------------------------------------------------
def update_job(status, start_date=None, end_date=None, output_log=None, error_log=None):
    logger.debug(f'Updating status: {status}')
    body={
        "status": status,
        "startDate": start_date.isoformat() if start_date is not None else None,
        "endDate": end_date.isoformat() if end_date is not None else None,
        "outputLog": output_log,
        "errorLog": error_log
    }
    # filter out the items where the value is None
    body = {key:value for key, value in body.items() if value is not None}
    try_request(requests.post, API_UPDATE_JOB, headers=headers, json_data=body)

# Charging thread ---------------------------------------------------
def start_charging_thread(stop_event: threading.Event):
    logger.debug('Start charging thread')
    def charge_thread():
        while not stop_event.is_set():
            res = try_request(requests.post, API_JOB_LIFECYCLE, headers=headers)
            if res is None:
                # stop looping, job is forced to terminate by try_request
                break
            j = res.json()
            if not j['succeeded']:
                # unable to validate job lifecycle due to known reason (e.g. insufficient funds, time/token limit exceeded)
                logger.error(f"Job lifecycle error! Response message: {j['message']} ")

            # @Todo: discuss if job should be charged for every computational minute that is started, i.e. if termination occurs
            # between two charge calls, should we charge for the minute that was started but not completed?
            stop_event.wait(runner_env.POSITRON_CHARGE_INTERVAL)
        logger.debug('Charging thread stopped')
    ct = threading.Thread(target=charge_thread)
    ct.start()
    return ct


# Termination thread ------------------------------------------------
def start_termination_thread(stop_event: threading.Event):
    logger.debug('Starting termination thread')
    def termination_thread():
        while not stop_event.is_set():
            res = try_request(requests.get, API_GET_JOB, headers=headers)
            if res is None:
                # stop looping, job is forced to terminate by try_request
                break
            j = res.json()
            if j['status'] == job_statuses['terminate']:
                logger.debug(f"Job status: {job_statuses['terminate']}, terminating job...")
                reasonExists = False
                if "meta" in j:
                    if "termination" in j["meta"]:
                        if "reason" in j["meta"]["termination"]:
                            reasonExists = True
                if reasonExists:
                    logger.warn(f"Termination reason: {j['meta']['termination']['reason']}")
                else:
                    logger.warn("Termination reason was not found.")
                terminate_running_job()
                global final_status; final_status = job_statuses["terminated"]
                break
            stop_event.wait(runner_env.POSITRON_CHECK_TERMINATION_INTERVAL)
        logger.debug('Termination thread stopped')

    tt = threading.Thread(target=termination_thread)
    tt.start()
    return tt


# Logging thread ----------------------------------------------------
def logging_thread(pipe, stream_name: str, stop_event: threading.Event):
    try:
        with pipe:
            while not stop_event.is_set():
                # Calling pipe.readline() will hang the process until a new line is available
                # however, if no new lines are available, we want to check if the thread should stop
                ready, _, _ = select.select([pipe], [], [], 0.1)
                if ready:
                    # TODO: Can I iterate over the pipe and process everything that's ready?
                    line = pipe.readline()
                    if not line:
                        break
                    if line.strip():
                        if stream_name == "stdout":
                            logger.info("job stdout: " + line.rstrip())
                        else:
                            logger.error("job stderr: " + line.rstrip())
                else:
                    time.sleep(0.01)
            # TODO: Do we need to check one more time for outstanding data in the pipe?
            logger.debug(f'Logging thread for: {stream_name} stopped')
    except Exception as e:
        logger.error(f'Logging thread for: {stream_name} stopped with exception: {e}')


# Terminate running job ---------------------------------------------
def terminate_running_job():
    global running_job
    # Signal termination
    if (running_job is not None) and (running_job.poll() is None):
        logger.warn("Forcefully terminating running job...")
        running_job.kill()
        running_job.wait()
        return_code = running_job.returncode
        logger.debug(f"Job process existed with return code: {return_code}")


# Retry API requests ------------------------------------------------
def try_request(request_func, url, retries=2, headers=None, json_data=None):
    for attempt in range(retries):
        try:
            response = request_func(url, headers=headers, json=json_data)
            response.raise_for_status()  # Raise an error for 4xx and 5xx status codes
            return response
        except requests.RequestException as e:
            logger.debug(f"Attempt {attempt + 1} failed: {e}")
            if attempt < retries - 1:
                logger.debug("Retrying...")
                time.sleep(2)  # Adding a delay before retrying
            else:
                logger.debug("Max retries reached, terminating job.")
                terminate_running_job()
                raise RobbieException(f"Failed to make request to {url} after {retries} attempts.")


'''_______________________________________________________________________________________
    
    RUN!
    ______________________________________________________________________________________
'''
def run():
    try:
        # 1. Get the job details
        job = get_job()

        if is_job_active():
            # 2. Update status to initializing
            update_job(status=job_statuses["initializing"], start_date=datetime.now(timezone.utc))

            # 3. Unpack workspace tar
            if is_job_active():
                unpack_workspace_from_s3()

                # 4. Run the job (including dependency install)
                if is_job_active():
                    run_job(job)

                    # 5. Update status to Storing
                    update_job(status=job_statuses['storing'])

                    # 6. Upload results
                    upload_results_to_s3()

        # 7. Get the completed job details
        job = get_job()
        # 8. Update status to success or error
        finish_job(job)

    except Exception as e:
        logger.error(f'Job stopped. An exception occurred: {str(e)}.')
        logger.end_of_logs_signal()
        raise e

        



from pydantic import BaseModel, field_validator
from typing import Dict, Optional, List, Union
import os
import copy
import yaml
import re
from pydantic_yaml import to_yaml_str
from positron_common.utils import get_version, _current_python_version
from positron_common.exceptions import RobbieException
from positron_common.user_config import user_config, REMOTE_FUNCTION_SECRET_KEY_NAME
from positron_common.constants import JOB_CONF_YAML_PATH, RUNTIME_ENV_PREFIX
from positron_common.cli.logging_config import logger
from positron_common.enums import JobRunType

class PositronJob(BaseModel):
    """
    The Job details as defined in the `python_job` from the `job_config.yaml` file.
    """
    name: Optional[str] = None
    funding_group_id: Optional[str] = None
    environment_id: Optional[str] = None
    image: Optional[str] = None # this is not written to the config file
    job_type: Optional[JobRunType] = None
    commands: Optional[List[str]] = None
    workspace_dir: Optional[str] = None
    max_tokens: Optional[int] = None
    max_time: Optional[int] = None
    env: Optional[Dict[str, str]] = None
    dependencies: Optional[str] = None
    conda_env: Optional[str] = None
    # below are values not writtent to the config file
    cluster: Optional[str] = None
    python_version: Optional[str] = _current_python_version()
    robbie_sdk_version: Optional[str] = get_version()
    image_selection: Optional[str] = None
    funding_selection: Optional[str] = None
    environment_selection: Optional[str] = None

    @field_validator('commands', mode='after')
    def ensure_non_empty(cls, commands):
        return commands if len(commands) else None

    @field_validator('env', mode='after')
    def ensure_env_is_dict(cls, v):
        if isinstance(v, dict):
            return v
        raise ValueError('env must be a dictionary')

    @field_validator('max_time', mode='before')
    def ensure_max_time_is_int(cls, max_time: Union[int, str, None]) -> Union[int, None]:
        return cls._max_time_to_minutes(max_time)

    @field_validator('max_tokens', mode='before')
    def ensure_max_tokens_is_int(cls, max_tokens: Union[int, str, None]) -> Union[int, None]:
        return cls._max_tokens_to_int(max_tokens)

    def create_runtime_env(self) -> Dict[str, str]:
        """
        Used on the client side to create the prefixed runtime environment variables
        to avoid conflicts with the local environment variables.
        """
        env: Dict[str, str] = {}
        # without env_prefix?? yes, but maybe not in this file, but part of the deploy.py file.
        env[REMOTE_FUNCTION_SECRET_KEY_NAME] = user_config.user_auth_token
        if not self.env:
            return env
        for key, value in self.env.items():
            if (value == ""):
                env_var = os.environ.get(key)
                if env_var is None:
                    raise ValueError(f"The env prop {key} is unset inside job_config.yaml and also unset in local env vars. Please set this value.")
                env[f'{RUNTIME_ENV_PREFIX}{key}'] = env_var
            else:
                env[f'{RUNTIME_ENV_PREFIX}{key}'] = value
        return env

    @staticmethod
    def _max_time_to_minutes(max_time: Union[int, str, None]) -> Union[int, None]:
        if not max_time:
            return None
        if isinstance(max_time, int):
            return max_time
        matches = re.search(r'^(\d+):(\d{2})$', max_time)
        if matches is None:
            raise ValueError(f'Invalid Job Config: Field "max_time" ({max_time}) must have the format "HH:MM" or be a positive integer')
        try:
            hours = int(matches.group(1))
            minutes = int(matches.group(2))
        except:
            raise ValueError(f'Invalid Job Config: Field "max_time" ({max_time}) must have the format "HH:MM" or be a positive integer')
        if minutes >= 60:
            raise ValueError('Invalid Job Config: Field "max_time" ({max_time}) has invalid minutes! Must be 0 <= minutes < 60!')
        return hours * 60 + minutes

    @staticmethod
    def _max_tokens_to_int(max_tokens: Union[int, str, None]) -> Union[int, None]:
        if not max_tokens:
            return None
        if isinstance(max_tokens, int):
            return max_tokens
        try:
            max_tokens = int(max_tokens)
        except:
            raise ValueError(f'Invalid Job Config: "max_tokens" ({max_tokens}) needs to be a positive integer.')
        if max_tokens <= 0:
            raise ValueError(f'Invalid Job Config: "max_tokens" ({max_tokens}) needs to be a positive integer.')
        return max_tokens

    def validate_values(self) -> None:
        errors = []
        if self.env and not validate_env_vars(self.env):
            errors.append('At least one of the environment variables provided is invalid')
        if errors:
            raise RobbieException(f'Invalid configuration. Errors: {errors}')
        return None
    
    def namestr(self, obj, namespace):
        return [name for name in namespace if namespace[name] is obj]

    def to_string(self, title: str = 'None'):
        message = f"""
- python_version: {self.python_version}
- robbie_sdk_version: {self.robbie_sdk_version}
- name: {self.name}
- funding_group_id: {self.funding_group_id}
    - funding_selection: {self.funding_selection}
- environment_id: {self.environment_id}
    - environment_selection: {self.environment_selection}
- image: {self.image}
    - image_selection: {self.image_selection}
- cluster: {self.cluster}
- dependencies: {self.dependencies}
- conda_env: {self.conda_env}
- job_type: {self.job_type}
- workspace_dir: {self.workspace_dir}
- max_tokens: {self.max_tokens}
- max_time: {self.max_time}
- env: {self.env}
- commands: {self.commands}"""

        if title:
            return f"========== {title} ==========\n{message}"
        else:
            return message
        

class PositronJobConfig(BaseModel):
    """
    The `job_config.yaml` schema class.
    """
    version: float
    python_job: PositronJob

    def write_to_file(this, filename: str = JOB_CONF_YAML_PATH):
        copy_of_config = copy.deepcopy(this)
        del copy_of_config.python_job.robbie_sdk_version
        del copy_of_config.python_job.python_version
        del copy_of_config.python_job.cluster
        del copy_of_config.python_job.image_selection
        del copy_of_config.python_job.funding_selection
        del copy_of_config.python_job.environment_selection
        
        with open(filename, 'w') as file:
            file.write(to_yaml_str(copy_of_config, exclude_unset=True))


def is_valid_key_value(keyvalue):
    """
    Validate that the key-value contains only alphanumeric characters, dashes, and underscores, and has no spaces.
    """
    return bool(re.match(r'^[\w-]+$', keyvalue))

def validate_env_vars(env_dict):
    """
    Validate the environment variables from the given dictionary.
    """
    valid = True
    for key, value in env_dict.items():
        if not is_valid_key_value(key):
            print(f"Invalid key (contains invalid characters or spaces): {key}")
            valid = False
        if value != "" and not is_valid_key_value(value):
            print(f"Invalid value (contains invalid characters or spaces): {value}")
            valid = False
    return valid

def merge_config(base_config: PositronJob, override_config: PositronJob) -> PositronJob:
    """
    Makes it easy to merge decorator configs on top of the YAML config.
    """
    update_data = override_config.dict(exclude_unset=True)
    updated_config = base_config.copy(update=update_data)
    return updated_config

def parse_job_config(config_path: str = 'job_config.yaml') -> Optional[PositronJob]:
    """
    Load the job configuration from the `job_config.yaml` file if it exists
    """
    
    if not os.path.exists(config_path):
        logger.debug('job_config.yaml file not found')
        return None

    try:
        # this happens sometimes and it causes the job to fail
        if os.path.getsize(config_path) == 0:
            logger.debug('job_config.yaml file is empty')
            return None
        with open(config_path, 'r') as job_config_file:
            job_config_dict = yaml.safe_load(job_config_file)
            job_config = PositronJobConfig(**job_config_dict)
            return job_config.python_job
    except Exception as e:
        print(f'Error loading job configuration! {str(e)}')
        return None

# if main then load and validate
if __name__ == "__main__":
    job_config = parse_job_config()
    if job_config:
        job_config.validate_values()
        print(job_config)


import os
from typing import Any
from .s3_client import get_s3_client
from .runner_env import runner_env
from .cloud_logger import logger

class CloudStorageClient:
    """Client for interacting with cloud storage. Abstracts S3 out."""
    def __init__(self):
        self.s3_client = get_s3_client()

    def download_file(self, s3_key: str, local_path: str):
        """Download a file from S3"""
        # ensure the local directory exists
        local_dir = os.path.dirname(local_path)
        if not os.path.exists(local_dir):
            os.makedirs(local_dir)
        try:
            self.s3_client.download_file(runner_env.S3_BUCKET, s3_key, local_path)
            logger.debug(f"Downloaded {s3_key} from S3 to {local_path}")
        except Exception as e:
            logger.error(f"Failed to download {s3_key} from S3: {e}")
            raise

    def get_object(self, s3_key: str) -> Any:
        """Get the contents of an S3 object into memory and return"""
        # TODO: may be used...
        try:
            response = self.s3_client.get_object(Bucket=runner_env.S3_BUCKET, Key=s3_key)
            return response['Body'].read()
        except Exception as e:
            logger.error(f"Failed to get {s3_key} from S3: {e}")
            raise

    def upload_file(self, local_path: str, s3_key: str):
        """Upload a file to S3"""
        bucket_name = runner_env.S3_BUCKET
        try:
            self.s3_client.upload_file(local_path, bucket_name, s3_key)
            logger.info(f"Uploaded {local_path} to S3 at {s3_key}")
        except Exception as e:
            logger.error(f"Failed to upload {local_path} to S3: {e}")
            raise

cloud_storage = CloudStorageClient()



import typer
from typing import Annotated
from .run_job import run
from .runner_env import runner_env
from .cloud_logger import logger

app = typer.Typer(help="Runs your job from the Positron Cloud environment.")

@app.command()
def hello():
    """
    Describes what you can do with the Positron Job Runner
    """
    logger.info('Hello, I am the Positron Job Runner')
    logger.info('Here is a list of thing I can help you with:')
    logger.info('- Run a job in the Positron Cloud')

@app.command()
def run_job(
    rerun: Annotated[bool, typer.Option(help='Enables rerunning a job')] = False,
):
    """
    Run the job from inside a Positron container.

    Example usage:
    $ positron_job_runner run_job
    """
    # @TODO: Validate that all necessary env variables are correctly set (JOB_ID...)

    logger.info('Running job in the Positron Cloud')
    logger.info(f'Job ID: {runner_env.JOB_ID}')
    runner_env.rerun = rerun

    logger.debug(f"Wrapper env: {runner_env}")

    try:
        run()
    except Exception as e:
        logger.exception(e)
        exit(0)


if __name__ == "__main__":
    app()



import dill

do_not_load = ['open', 'os', 'do_not_load', 'remote_load_pickle', 'remote_local_pickle']

LOCAL_PICKLE = 'cell.pkl'
REMOTE_PICKLE = 'cell_result.pkl'

def remote_load_pickle():
    file = open(LOCAL_PICKLE, 'rb')
    unpickler = dill.Unpickler(file)
    user_ns = unpickler.load()
    file.close()
    if user_ns is not None:
        for key, value in user_ns.items():
            if key not in do_not_load and (key not in globals() or not callable(globals()[key])):
                globals()[key] = value

def remote_local_pickle(user_ns: dict):
    ns_copy = user_ns.copy()
    # baditems = dill.detect.baditems(user_ns)
    # for baditem in baditems:
    ns_copy = {key: value for key, value in ns_copy.items() if key not in do_not_load}
    
    file = open(REMOTE_PICKLE, 'wb')
    protocol = dill.settings['protocol']
    pickler = dill.Pickler(file, protocol)
    pickler._byref = False   # disable pickling by name reference
    pickler._recurse = False # disable pickling recursion for globals
    pickler._session = True  # is best indicator of when pickling a session
    pickler._first_pass = True
    pickler.dump(ns_copy)
    file.flush()
    file.close()

remote_load_pickle()
    
### {cell}

remote_local_pickle(globals())


from positron_job_runner.runner_env import runner_env
from positron_job_runner.cloud_logger import logger
from positron_job_runner.cloud_storage import cloud_storage
from positron_common.deployment.meta_data import MetaData
from positron_common.deployment.serializer import Serializer

class RemoteFunction():
  """Fetches the function to execute from cloud storage."""

  run_cmd: str = "python -m positron_job_runner.lil_byter"
  s3_base_path = f"{runner_env.JOB_OWNER_EMAIL}/{runner_env.JOB_ID}"

  def setup(self):
    logger.info("Fetching function to execute...")

    meta_contents = cloud_storage.get_object(f"{self.s3_base_path}/function/metadata.json")
    metadata = MetaData.from_json(meta_contents)

    # TODO: use positron_common enum s3_constants
    self._get_resource('function.pkl', metadata.sha256_hash.get('func_hash'))
    self._get_resource('args.pkl', metadata.sha256_hash.get('args_hash'))
    self._get_resource('kwargs.pkl', metadata.sha256_hash.get('kwargs_hash'))

    logger.info("Function successfully retrieved and validated.")

  def _get_resource(self, resource_name: str, resource_hash: str):
    # TODO: use positron_common enum s3_constants
    contents = cloud_storage.get_object(f"{self.s3_base_path}/function/{resource_name}")
    computed_hash = Serializer.compute_hash(contents, secret_key=runner_env.REMOTE_FUNCTION_SECRET_KEY)
    if computed_hash != resource_hash:
      raise Exception(f"Integrity check for: {resource_hash} failed! Ensure your User Auth Token is up to date and no one has access to your cloud storage.")

    # Write to disk so subprocess can run it.
    with open(f"{runner_env.JOB_CWD}/{resource_name}", "wb") as f:
      f.write(contents)

    logger.info(f"Downloaded {resource_name}")


if __name__ == "__main__":
  rf = RemoteFunction()
  rf.setup()



import hmac
import hashlib
from typing import Any, Tuple
import cloudpickle

from positron_common.exceptions import SerializationException, DeserializationException

class Serializer:
    """Serializer class for serializing and deserializing objects with CloudPickle."""


    @staticmethod
    def serialize(obj: Any) -> Any:
        try:
            bytes_to_upload = cloudpickle.dumps(obj)
            return bytes_to_upload
        except Exception as e:
            raise SerializationException("Object is not serializable") from e

    @staticmethod
    def deserialize(bytes_to_deserialize: bytes) -> Any:
        try:
            return cloudpickle.loads(bytes_to_deserialize)
        except Exception as e:
            raise DeserializationException(
                "Error when deserializing bytes: {}".format(repr(e))
            ) from e

    @staticmethod
    def compute_hash(buffer: bytes, secret_key: str) -> str:
        return hmac.new(secret_key.encode(), msg=buffer, digestmod=hashlib.sha256).hexdigest()



import os
from dataclasses import dataclass
from dotenv import load_dotenv
from positron_common.user_config import REMOTE_FUNCTION_SECRET_KEY_NAME
from positron_common.constants import temp_path, RUNTIME_ENV_PREFIX
from typing import Union

load_dotenv()

# Disable stdout buffering
os.environ['PYTHONUNBUFFERED'] = '1'
os.environ['POSITRON_CLOUD_ENVIRONMENT'] = '1'

@dataclass
class RunnerEnv():
  """
  The environment variables and other global vars required by the Runner.
  """
  SYSTEM_AUTHENTICATION_KEY: str = os.getenv('SYSTEM_AUTHENTICATION_KEY', "")
  REGION: str = os.getenv("REGION", "us-west-2")
  AWS_JOB_LOG_GROUP_NAME: str = os.getenv("AWS_JOB_LOG_GROUP_NAME", "")
  API_ENDPOINT: str = os.getenv("API_ENDPOINT", "")
  POSITRON_CLI_ENV: str = os.getenv("POSITRON_CLI_ENV", "development")
  POSITRON_CHARGE_INTERVAL: int = int(os.getenv('POSITRON_CHARGE_INTERVAL', 60))
  POSITRON_STDOUT_INTERVAL: int = int(os.getenv('POSITRON_STDOUT_INTERVAL', 2))
  POSITRON_CHECK_TERMINATION_INTERVAL: int = int(os.getenv('POSITRON_CHECK_TERMINATION_INTERVAL', 10))
  RUNNER_CWD: str = os.getenv('RUNNER_CWD', os.path.join(os.getcwd(), f'{temp_path}/job-controller'))
  ROBBIE_CLI_PATH: str = os.getenv('ROBBIE_CLI_PATH', '$HOME/python-decorator')
  APP_NAME: str = os.getenv('APP_NAME', 'robbie')
  LOG_LEVEL: str = os.getenv('LOG_LEVEL', "DEBUG") # see logging.getLevelNameMapping()
  JOB_ID: Union[str, None] = os.getenv('JOB_ID', None)
  JOB_CWD: str = os.getenv('JOB_CWD', os.path.join(os.getcwd(), f'{temp_path}/job-execution'))
  JOB_USER_CWD: str = os.getenv('JOB_USER_CWD', os.path.join(os.getcwd(), f'{temp_path}'))
  JOB_USER: Union[str, None] = os.getenv('JOB_USER', None)
  JOB_OWNER_EMAIL: Union[str, None] = os.getenv('JOB_OWNER_EMAIL', None)
  S3_BUCKET: str = os.getenv('S3_BUCKET', 'positron-dev-workspaces')
  ROBBIE_RUNNER_AWS_ACCESS_KEY_ID: Union[str, None] = os.getenv('ROBBIE_RUNNER_AWS_ACCESS_KEY_ID')
  ROBBIE_RUNNER_AWS_SECRET_ACCESS_KEY: Union[str, None] = os.getenv('ROBBIE_RUNNER_AWS_SECRET_ACCESS_KEY')
  REMOTE_FUNCTION_SECRET_KEY: str = os.getenv(REMOTE_FUNCTION_SECRET_KEY_NAME, "")
  rerun: bool = False

  def env_without_runner_env(self):
    """
    Will blacklist env vars we're explicitly expecting to be
    set by the infra and used by the Runner.
    """
    env = {}
    # if the var is prefixed, then it was passed by the user and should be included
    for k, v in os.environ.items():
      if k.startswith(RUNTIME_ENV_PREFIX):
        env[k[len(RUNTIME_ENV_PREFIX):]] = v

    # Manually set required vars.
    env['HOME'] = f"/home/{self.JOB_USER}"
    env['PWD'] = f"{self.JOB_USER_CWD}"
    env['PATH'] = f"{self.JOB_USER_CWD}/venv/bin:/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
    env['VIRTUAL_ENV'] = f"{self.JOB_USER_CWD}/venv"
    env['PYTHONUNBUFFERED'] = '1'
    env['POSITRON_CLOUD_ENVIRONMENT'] = '1'

    # Explicitly set as this is just the users API key and will be used to calc integrity hash
    # in the subprocess if this is the remote function.
    env[REMOTE_FUNCTION_SECRET_KEY_NAME] = self.REMOTE_FUNCTION_SECRET_KEY
    return env

runner_env = RunnerEnv()



import os

APP_NAME = "robbie"
APP_HOME_DIR = os.path.expanduser("~/.robbie")
temp_path: str = '.robbie'
JOB_CONF_YAML_PATH = "./job_config.yaml"

NERC_CLUSTER = "NERC"
EKS_CLUSTER = "EKS"

# Used to prefix environment variables that are passed to the runtime environment
RUNTIME_ENV_PREFIX: str = '__RUNTIME_'

# 50MB
FILE_SIZE_THRESHOLD = 50 * 1024 * 1024



# TODO: move this file to positron_common
import boto3
from botocore.client import BaseClient
from .runner_env import runner_env

def get_s3_client() -> BaseClient:
    # Get the ROBBIE_RUNNER-prefixed credentials if they exist - this avoids breaking dev workflows
    robbie_access_key_id = runner_env.ROBBIE_RUNNER_AWS_ACCESS_KEY_ID
    robbie_secret_access_key = runner_env.ROBBIE_RUNNER_AWS_SECRET_ACCESS_KEY

    # Initialize S3 client
    if robbie_access_key_id and robbie_secret_access_key:
        # If the ROBBIE_RUNNER-prefixed credentials exist, use them
        session = boto3.Session(
            aws_access_key_id=robbie_access_key_id,
            aws_secret_access_key=robbie_secret_access_key
        )
        return session.client('s3', region_name=runner_env.REGION)
    # If ROBBIE_RUNNER-prefixed credentials don't exist, use default credentials from environment
    return boto3.client('s3',region_name=runner_env.REGION)



import os
import typer
import copy
from typing_extensions import Annotated
from typing import Optional
from positron_common.deploy import command_runner_deploy
from positron_common.config import parse_job_config, PositronJob
from positron_common.cli_args import args
from positron_common.cli.console import console, ROBBIE_DEFAULT, ROBBIE_BLUE
from positron_common.constants import JOB_CONF_YAML_PATH
from positron_common.image_name import get_auto_image_name_and_cluster
from positron_common.cli.interactive import prompt_and_build_positron_job_config
from positron_common.cli.logging_config import logger
from positron_common.enums import JobRunType
from positron_common.observability.main import track_command_usage
from positron_cli.download import filename_is_valid
from positron_common.job_api.funding_envs_images import *
from positron_common.job_api.validate_user_auth_token import is_auth_token_valid
from positron_common.user_config import user_config
from positron_cli.login import login
from pipreqs.pipreqs import init
from prompt_toolkit import prompt
from prompt_toolkit.styles import Style



@track_command_usage("run")
def run(
  # the following options control the job configuration
  name_arg: Annotated[str, typer.Option("--name", help="Name of the run")] = None,
  funding_arg: Annotated[str, typer.Option("--funding_group_id", help="Specify the team to run the job on.")] = None,
  environment_arg: Annotated[str, typer.Option("--environment_id", help="Specify the hardware to run the job on.")] = None,
  image_arg: Annotated[str, typer.Option("--image", help="Specify the image to run the job on or 'auto-select'.")] = None,
  commands_arg: Annotated[Optional[str], typer.Argument(help='String of shell commands, a .py file, or a .ipynb file.')] = None,
  auto_capture_deps: Annotated[bool, typer.Option("--auto-dep", help='(EXPERIMENTAL) Automatically resolve Python dependencies.')] = None,
  # the following options control how the job is run
  tail: Annotated[bool, typer.Option("--tail", help='Tail the run\'s stdout back to your CLI.')] = False,
  status: Annotated[bool, typer.Option("--s", help='Monitor run status in your CLI.')] = False,
  skip_prompts: Annotated[bool, typer.Option("--y", help='Bypass the prompts and execute the run immediately.')] = False,
  interactive: Annotated[bool, typer.Option("--i", help="Interactively choose your run configuration.")] = False,
  download: Annotated[str, typer.Option("--download", help="Download a results <file> to your local machine. Specify 'all' for the complete list.")] = None,
  path: Annotated[str, typer.Option("--path", help="Local directory where the downloaded files will be stored.")] = None,
  create_only: Annotated[bool, typer.Option("--create-only", help="Only create the run, do not execute it. [Robbie internal use only]")] = False,
) -> None:
    """
    Run shell commands as a batch job in Robbie

    There are three possible scenarios:
    1. User types: robbie run - commands are run from the job_config.yaml file
    2. User types: robbie run "command" - commands override the job_config.yaml file
    3. User types: robbie run --i - interactive mode, user is prompted for all the options and a job_config.yaml file is created/overwritten

    In scenerio 1 and 2, user's can pass in funding groups and enviroments as arguments to override the job_config.yaml file.

    """

    logger.debug(f"""========== run() arguments ==========
    - name_arg: {name_arg}
    - funding_arg: {funding_arg}
    - environment_arg: {environment_arg}
    - image_arg: {image_arg}
    - commands_arg: {commands_arg}
    - auto_capture_deps: {auto_capture_deps}
    - tail: {tail}
    - status: {status}
    - skip_prompts: {skip_prompts}
    - interactive: {interactive}
    - download: {download}
    - path: {path}
    - create_only: {create_only}""")


    # This check eliminates a extra step if the user happens to run robbie run and the API key is not valid
    if not user_config.user_auth_token or not is_auth_token_valid():
        # there is no auth token set in .robbie/config.yaml or the token is invalid/expired in the backend
        console.print('[red]Sorry, your API is invalid, please login.')
        login()

    # we can either stream or monitor status but not both at the same time
    if tail and status:
        console.print('[red]Error: Choose either the -logs and -s option.')
        return
    
    if (download and not (tail or status)):    
        console.print('[red]Error: The --download option can only be used with the --tail or --s option.')
        return
    
    if download and not filename_is_valid(download):
        console.print('[red]Error: Please specify a valid file name or "all" to download all files.')
        return
    
    if path and not download:
        console.print('[red]Error: The --path option can only be used with the --download option.')
        return
    
    if path and not os.path.exists(path):
        console.print('[red]Error: The path you specified is not valid.')
        return

    # initialize the argument singleton
    args.init(
        name=name_arg,
        stream_stdout=tail,
        skip_prompts=skip_prompts,
        monitor_status=status,
        commands_to_run=commands_arg,
        interactive=interactive,
        create_only=create_only,
        download=download,
        local_path=path,
        auto_capture_deps=auto_capture_deps
    )

    # first-level sanity checks
    if commands_arg and interactive:
        console.print("[red]Sorry: Please specify command line or use the interactive mode.")
        return
    
    if interactive and (funding_arg or environment_arg or image_arg):
        console.print("[red]Sorry: You can't specify funding, environments, or images in interactive mode.")
        return

    # Read in the job_config.yaml file if it exists
    yaml_job_config = None
    if os.path.exists(JOB_CONF_YAML_PATH) and os.path.getsize(JOB_CONF_YAML_PATH) != 0:
        console.print(f'Found run configuration file (reading...): {JOB_CONF_YAML_PATH}')
        yaml_job_config = parse_job_config(JOB_CONF_YAML_PATH)
        if yaml_job_config:
            if yaml_job_config.funding_group_id:
                yaml_job_config.funding_selection = "From job_config.yaml"
            if yaml_job_config.environment_id:
                yaml_job_config.environment_selection = "From job_config.yaml"
            if yaml_job_config.image:
                yaml_job_config.image_selection = "From job_config.yaml"
        else:
            console.print('[red]Error parsing job_config.yaml file. Disregarding it.')
            

    """
    There are three possible scenarios:
    1. User types: robbie run - commands are run from the job_config.yaml file
    2. User types: robbie run "command" - commands override the job_config.yaml file
    3. User types: robbie run --i - interactive mode, user is prompted for all the options and a job_config.yaml file is created

    There are three important variables in the code:
    - yaml_job_config - this is the PositronJob object created from the job_config.yaml file
    - job_config_to_run - this is the PositronJob object that will be passed to deploy to run the job
    - add_args_to_run_job_config - this is a flag that is set for scenerio 1 and 2 to merge the passed arguments into to the job_config_to_run object
    """
    add_args_to_run_job_config = False
    # 
    # Scenerio 1: User types: "robbie run" - commands are run from the job_config.yaml file
    #
    if not commands_arg and not interactive:
        logger.debug("@@@@@@@@@@@   Scenerio 1: User types: robbie run")
        if not yaml_job_config:
            console.print('[red]Error: No job_config.yaml file found. Please specify commands to run.')
            return
        if not yaml_job_config.commands:
            console.print('[red]Error: No commands found in job_config.yaml file.')
            return
        job_config_to_run = yaml_job_config
        add_args_to_run_job_config = True

    # 
    # Scenerio 2. User types: robbie run "command" - commands override the job_config.yaml file if it exists
    #
    if commands_arg:
        logger.debug("@@@@@@@@@@ Scenerio 2: User types: robbie run 'command'")
        # Is there a job_config.yaml file?
        if yaml_job_config:
            if yaml_job_config.commands:
                console.print('[yellow]Overriding commands in existing job_config.yaml file.')
            job_config_to_run = yaml_job_config
        else:
            console.print('No job_config.yaml file found, using defaults.')
            job_config_to_run = PositronJob()
        
        # Add the commands to the job_config
        job_config_to_run.commands = []
        job_config_to_run.commands.append(commands_arg)

        add_args_to_run_job_config = True
        
    # now lets deal with the arguments for the first two scenarios
    if add_args_to_run_job_config:
        logger.debug("Adding arguments to job_config")
        job_config_to_run = _merge_from_yaml_and_args(job_config_to_run, funding_arg, environment_arg)
        if not job_config_to_run:
            logger.debug('[red]Error: Unable to merge yaml and arguments.')
            return
        
        # now lets sort out the image
        if job_config_to_run.image:
            # there is one in the job_config.yaml file
            if image_arg:
                logger.debug(f'[yellow] Notice: overriding image job_config.yaml image: {job_config_to_run.image}, setting image to: {image_arg}')
                job_config_to_run.image = image_arg
                job_config_to_run.image_selection = "Overriden by argument"
            else:
                logger.debug(f'Using the image: {job_config_to_run.image} found in job_config.yaml. To override, please use the --image option.')
                job_config_to_run.image_selection = "From job_config.yaml"
    
            if job_config_to_run.image == "auto-select":
                # auto select the image and cluster
                job_config_to_run.image, job_config_to_run.cluster = get_auto_image_name_and_cluster(job_config_to_run.funding_group_id, job_config_to_run.environment_id)
                job_config_to_run.image_selection = "Automatically Selected"
    # 
    # Scenerio 3. User types: robbie run --i - interactive mode, user is prompted for all the options and a job_config.yaml file is created or modified
    #
    if interactive:
        logger.debug("@@@@@@@@@ Scenerio 3: User types: robbie run --i")
        preserve = False
        if yaml_job_config and yaml_job_config.commands:
            response = prompt('You are about to override your job_config.yaml, do you want preserve the enviroment variables and commands [[y]/n]:', style=Style.from_dict({'prompt': 'yellow'}))
            if response in ["y", "yes", "Yes", "Y", ""]:
                preserve = True        
                       
        # lets prompt the user 
        console.print(f"Please follow the prompts to configure your run ([{ROBBIE_DEFAULT}][] = default[/{ROBBIE_DEFAULT}], <tab> for options)")
        captured_job_config = prompt_and_build_positron_job_config(preserve)
        if captured_job_config == None:
            console.print(f"[red]Sorry, failed to create a file {JOB_CONF_YAML_PATH}")
            return
        
        logger.debug(captured_job_config.python_job.to_string("Interactive cfg"))
        # keeping the previous commands and envs from the job_config.yaml file
        if preserve:
            if yaml_job_config.env:
                # this prevents a null env from being copied over
                captured_job_config.python_job.env = yaml_job_config.env
            
            captured_job_config.python_job.commands = yaml_job_config.commands

        if captured_job_config.python_job.commands == None:
            console.print("[red]Error: You did not specify any commands to run.")
            return
        captured_job_config.write_to_file(filename=JOB_CONF_YAML_PATH)
        job_config_to_run = captured_job_config.python_job

    # Handle auto capturing dependencies for non-Conda environments
    # this comes in two forms:
    # 1. User types: robbie run --auto-dep
    # 2. job_config.yaml file has dependencies: auto-capture
    overwrite_reqs = False
    if auto_capture_deps or (yaml_job_config and yaml_job_config.dependencies == "auto-capture"):
        logger.debug("Auto-capturing dependencies")
        if os.path.exists("./requirements.txt"):
            if args.skip_prompts == False:
                overwrite_reqs = prompt('A requirement.txt file already exist, do you want to overwrite it? [y/N]:', style=Style.from_dict({'prompt': 'yellow'}))
                if overwrite_reqs in ["no", "n", "No", "N", ""]:
                    console.print("[yellow]See you soon![/yellow]")
                    return
            else:
                overwrite_reqs = True
            
        # arguments for pipreqs package - https://github.com/bndr/pipreqs
        pipreq_args = {
            "<path>": ".",
            "--print": False,
            "--savepath": None,
            "--pypi-server": None,
            "--proxy": None,
            "--use-local": False,
            "--diff": None,
            "--clean": None,
            "--mode": None,
            "--scan-notebooks": True,
            "--force": False
        }
        if overwrite_reqs:
            pipreq_args["--force"] = True

        console.print("Analyzing Python (.py) and Notebook (.ipynb) files for dependencies...", style=ROBBIE_BLUE)
        init(pipreq_args) # this is the pipreqs call    
        console.print("[green]✔[/green] Dependency analysis complete, requirements.txt file created.")
        job_config_to_run.dependencies = "./requirements.txt"
        # hack to remove robbie from the generated requirements.txt file
        _remove_package_from_requirements('robbie')

    job_config_to_run.job_type = JobRunType.BASH_COMMAND_RUNNER
    logger.debug("Calling command_runner_deploy")
    command_runner_deploy(job_config_to_run)


def _merge_from_yaml_and_args(input_job_config: PositronJob, funding_arg: str, environment_arg: str) -> PositronJob:
    """
    Merge the job_config.yaml file with the command line arguments for funding, and environment.
    Ensure that environment is in the funding group.

    Behavior:
    - Command line arguments take precedence over the job_config.yaml file
    - Command line arguments are not memorized in the job_config.yaml file 

    """
    if not input_job_config:
        logger.debug("No job config to merge.")
        return None
    
    logger.debug(input_job_config.to_string("Entering _merge_from_yaml_and_args()"))
    
    return_config = copy.deepcopy(input_job_config)
    logger.debug(f"return_config: {id(return_config)} created from input_job_config: {id(input_job_config)}")

    # Is there a funding group in the input config?
    if input_job_config.funding_group_id:
        logger.debug(f"return_config.funding_group_id: already set: {return_config.funding_group_id}")
        # ok the correct funding group is already in return_config.funding_group_id
        pass

    # Is there a funding group argument?
    if funding_arg:
        logger.debug(f"Setting or overriding return_config.funding_group_id: {return_config.funding_group_id} <-- funding_arg: {funding_arg}")
        return_config.funding_selection = "Overridden by argument"
        return_config.funding_group_id = funding_arg
        
    # pre-fetch the funding sources info and def env
    fs = list_funding_sources()
    if len(fs) == 0:
        logger.debug("No funding sources found.")
        return None

    def_env_id = None
    # Have we identified a funding group yet?
    if return_config.funding_group_id:
        # Find the default funding source (personal)
        logger.debug(f"Yes, we have a funding group: {return_config.funding_group_id}")
        if not return_config.environment_id:
            for _, val in fs.items():
                if (val[FS_ID] == return_config.funding_group_id):
                    def_env_id = val.get(FS_DEF_ENV_ID)
            if not def_env_id:
                console.print(f"[bold red]Can't get the default environment for the funding group: {return_config.funding_group_id}")
                return None   
            if not return_config.environment_id:
                logger.debug(f"setting return_config.environment_id: {return_config.environment_id} <-- def_env_id: {def_env_id}")
                return_config.environment_id = def_env_id
                return_config.environment_selection = "Default"
        else:
            if not _env_in_funding_group(return_config.environment_id, return_config.funding_group_id):
                console.print(f"[bold red] Sorry, the environment in your job_config.yaml: {return_config.environment_id} is not the funding group: {return_config.funding_group_id}")
                return None
    else: 
        # No funding group id, let's get the user's personal and defaut environment
        logger.debug("Still no funding group id, let's get the user's personal one.")
        for _, val in fs.items():
            if (val[FS_TYPE] == FS_PERSONAL_TYPE):
                return_config.funding_group_id = val.get(FS_ID)
                def_env_id = val.get(FS_DEF_ENV_ID)
                logger.debug(f"Setting return_config.funding_group_id to PERSONAL: {return_config.funding_group_id}, fetched def_env_id: {def_env_id}")
        if return_config.funding_group_id:
            return_config.funding_selection = "Default"
        else:
            console.print("[bold red]Can't get the default funding source (personal).")
            return None
        if not def_env_id:
            console.print("[bold red]Can't get the default environment for the personal funding source.")
            return None
        # the user specified an environment in the job_config.yaml file
        if return_config.environment_id:
            # just for fun, lets check if its in the funding group
            if not _env_in_funding_group(return_config.environment_id, return_config.funding_group_id):
                console.print(f"[bold red] Sorry, the environment in your job_config.yaml: {return_config.environment_id} is not the funding group: {return_config.funding_group_id}")
                return None
        else:
            logger.debug(f"setting return_config.environment_id: {return_config.environment_id} <-- def_env_id: {def_env_id}")
            return_config.environment_id = def_env_id
            return_config.environment_selection = "Default from Funding Source"
        
    # At this point we should have a funding group id and a default environment id
    if environment_arg:
        # check to make certain the environment argument is in the funding group
        if not _env_in_funding_group(environment_arg, return_config.funding_group_id):
            console.print(f"[bold red] Sorry, the environment you passed as an argument: {environment_arg} is not the funding group: {return_config.funding_group_id}")
            return None
        logger.debug(f"setting return_config.environment_id: {return_config.environment_id} <-- environment_arg: {environment_arg}")
        return_config.environment_id = environment_arg
        return_config.environment_selection = "Overridden by argument"

    logger.debug(return_config.to_string("Exiting _merge_from_yaml_and_args()"))

    return return_config

def _env_in_funding_group(env_id, funding_group_id):
    """ sanity check a environment id is in the funding group """
    envs = list_environments(funding_group_id)
    for _, val in envs.items():
        if (val[FS_ID] == env_id):
            return True
    return False

def _remove_package_from_requirements(package_name, file_path='./requirements.txt'):
    """ this is hack when the robbie package gets auto-added to the requirements.txt file in the auto-capture dependencies mode """
    # Read the file
    with open(file_path, 'r') as file:
        lines = file.readlines()

    # Write back all lines except the one with the package name
    with open(file_path, 'w') as file:
        for line in lines:
            if not line.strip().startswith(package_name):
                file.write(line)










import json
import logging
import time
import os
from .runner_env import runner_env
from logging import _nameToLevel

LOGGER_NAME = f"{runner_env.APP_NAME}.job_runner"
JSON_LOGS=False
JSON_LOGS_PRETTY=False
JSON_LOGS_FILE_PATH=False

class _UTCFormatter(logging.Formatter):
    """Class that overrides the default local time provider to GMT in log formatter."""
    converter = time.gmtime


    def format(self, record):
        if not JSON_LOGS:
            return super(_UTCFormatter, self).format(record)

        log_record = {
            "asctime": time.strftime("%Y-%m-%d %H:%M:%S:%MS", self.converter(record.created)),
            "name": record.name,
            "level": record.levelname,
            "message": record.getMessage(),
        }

        if JSON_LOGS_FILE_PATH:
            relative_path = os.path.relpath(record.pathname)
            # Append the line number to the relative path
            relative_path_with_lineno = f"{relative_path}:{record.lineno}"
            log_record["pathname"] = relative_path_with_lineno

        if JSON_LOGS_PRETTY:
            return json.dumps(log_record, indent=4)
        return json.dumps(log_record)

def get_logger():
    logger = logging.getLogger(LOGGER_NAME)
    if len(logger.handlers) == 0:
        logger.setLevel(_nameToLevel.get(runner_env.LOG_LEVEL))
        handler = logging.StreamHandler()
        formatter = _UTCFormatter("%(asctime)s %(name)s %(levelname)-8s %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        # don't stream logs with the root logger handler
        logger.propagate = 0

    return logger



from requests import Response
from positron_common.cli.logging_config import logger
from positron_common.job_api.error_model import ResponseErrorModel

class RemoteCallException(Exception):
    """
    Parent exception class for Robbie package
    """
    user_friendly_message: str
    reason: str
    additional_help: str
    def __init__(self, user_friendly_message: str = "", reason: str = "", additional_help: str = ""):
        self.user_friendly_message = user_friendly_message
        self.reason = reason
        self.additional_help = additional_help

    @staticmethod
    def from_request(response: Response):
        try:
            error = ResponseErrorModel(**response.json())
            logger.debug(f"Error parsing request json: {error}")
            return RemoteCallException(
                user_friendly_message=error.userFriendlyErrorMessage,
                reason=error.message,
            )
        except Exception:
            logger.debug("Error parsing request json", exc_info=True)
            return RemoteCallException("Sorry, an error occurred! If the problem continues, reach out to our support team for help.\nEmail: support@robbie.run")
        
class RobbieException(Exception):
    """
    Parent exception class for Robbie package
    """
    pass

class SerializationException(RobbieException):
    """
    Exception raised when serialization errors occur.
    """
    pass

class DeserializationException(RobbieException):
    """
    Exception raised when deserialization errors occur.
    """
    pass

class DecoratorException(RobbieException):
    """
    Raised when decorator related errors are happening
    """
    pass

class RemoteFunctionException(Exception):
    """
    This is a wrapper exception to allow bubbling up of exceptions happening in a remote function call
    """
    rf_exception: Exception
    def __init__(self, e: Exception):
        self.rf_exception = e



from functools import wraps
import argparse
import os
from positron_common.cli_args import args as cli_args
from positron_common.config import PositronJob, parse_job_config, merge_config
from positron_common.deployment.deploy import Deploy
from positron_common.deployment.stored_function import StoredFunction
from positron_common.job_api.validate_user_auth_token import is_auth_token_valid
from positron_common.user_config import user_config
from positron_common.cli.console import console
from positron_common.cli.logging_config import logger, set_log_level
from positron_common.enums import JobRunType
from positron_common.utils import _exit_by_mode, _nb
from robbie.notebook_cell_ui import notebook_cell_ui
from positron_cli.login import login

def remote(**parameters):
    """
    Decorator to deploy a function in Robbie. 
    This works in both a Jupyter notebook and a Python script.

    You can pass argumments to the decorator to customize the deployment in two ways:
    - Command Line Arguments when running from the commmand line - e.g. python my_script.py --tail
        Supported arguments:
            --tail: bool
            --loglevel: str [CRITICAL,FATAL,ERROR, WARNING, INFO, DEBUG, NOTSET]
            --create-only: bool
            --results-from-job-id: str (job_id )
    - Decorator Aguments in a Python script or Jupyter Notebook when decorating a function - e.g. @remote(tail=True)
        Supported arguments:
            - funding_group_id: str
            - environment_id: str
            - image: str
            - tail: bool    
            - loglevel: str
            - create_only: bool
            - chooser_ui: bool (Enable user to choose job config via a small GUI in the cell - Used in Jupyter Notebook only)

    """

    # Parse command line arguments
    parser = argparse.ArgumentParser(description = "A decorator to handle deploying running your function in the cloud")
    parser.add_argument('--tail', action='store_true', help='Stream the stdout from Positron Cloud back to your cli', dest='stream_stdout', default=False)
    parser.add_argument('--loglevel', help='Set the logging level [CRITICAL,FATAL,ERROR, WARNING, INFO, DEBUG, NOTSET]', dest='loglevel')
    parser.add_argument('--create-only', action='store_true', help='Create the job but do not run it.', dest='create_only')
    parser.add_argument('--results-from-job-id', help='Fetch results and return from decorated function.', dest='results_from_job_id')
    positron_args, job_args = parser.parse_known_args()

    if positron_args.loglevel:
        set_log_level(positron_args.loglevel)

    logger.debug("========positron_args========")
    logger.debug(positron_args) 
    logger.debug("========job_args========")
    logger.debug(job_args)

    # Jupyter Support - Default out the cli_args to run remote always with no prompting
    if not cli_args.is_init:
        cli_args.init(
            local=False,
            deploy=True,
            stream_stdout=positron_args.stream_stdout,
            job_args=job_args,
            create_only=positron_args.create_only,
            results_from_job_id=positron_args.results_from_job_id,
            skip_prompts=True,
        )

    chooser_ui_flag = False
    # enable  and tail function parameters but remove them before passing to PositronJob config    
    if "loglevel" in parameters:
        set_log_level(parameters["loglevel"])
        del parameters["loglevel"]
    if "tail" in parameters:
        cli_args.stream_stdout = parameters["tail"]
        del parameters["tail"]
    if "create_only" in parameters:
        cli_args.create_only = parameters["create_only"]
        del parameters["create_only"]
    if "results_from_job_id" in parameters: 
        cli_args.results_from_job_id = parameters["results_from_job_id"]
        del parameters["results_from_job_id"]
    if "chooser_ui" in parameters:
        chooser_ui_flag = True
        del parameters["chooser_ui"]

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            logger.debug("Running decorator")

            # Check this first to ensure we don't deploy
            if os.getenv('POSITRON_CLOUD_ENVIRONMENT', False):
                logger.debug("Running function locally")
                return func(*args, **kwargs)

            # This check eliminates a extra step if the user happens to run robbie run and the API key is not valid
            if not user_config.user_auth_token or not is_auth_token_valid():
                # there is no auth token set in .robbie/config.yaml or the token is invalid/expired in the backend
                console.print('[red]Sorry, your API is invalid, please login.')
                logger.debug("Sorry, your API is invalid, please login.")
                login()

            if cli_args.results_from_job_id:
                stored_function = StoredFunction(func, args, kwargs)
                stored_function.set_job_id(cli_args.results_from_job_id)
                secret_key = user_config.user_auth_token if user_config.user_auth_token else ""
                stored_function.load_and_validate_results(hmac_key=secret_key)
                return stored_function.result


            # get decorator arguments
            job_config = None
            nonlocal chooser_ui_flag
            
            if chooser_ui_flag:
                if _nb: # we are in a notebook
                    job_config = notebook_cell_ui()
                    if job_config == None:
                        console.print("[red] User interrupted.")
                        return
                else:
                    console.print("[yellow]Warning: The 'chooser_ui' is only supported in Jupyter Notebooks. Please remove this argument.[/yellow]")
                    logger.warning("[yellow]Warning: The 'chooser_ui' is only supported in Jupyter Notebooks. Please remove this argument.[/yellow]")
                    _exit_by_mode(1)
            else:
                job_config_decorator = PositronJob(**parameters)
                # track where the parameters come from so we can display to the user later
                if job_config_decorator.funding_group_id:
                    job_config_decorator.funding_selection = "Passed as argument to @remote decorator"
                if job_config_decorator.environment_id:
                    job_config_decorator.environment_selection = "Passed as argument to @remote decorator"
                if job_config_decorator.image:
                    job_config_decorator.image_selection = "Passed as argument to @remote decorator"

                logger.debug(job_config_decorator.to_string("job_config_decorator (arguments passed into remote function)"))

                job_config = job_config_decorator


                # use job yaml as base if it exists
                job_config_yaml = parse_job_config()
                if job_config_yaml:
                    job_config_yaml.print("job_config_yaml found")
                    if job_config_yaml.funding_group_id:
                        job_config_yaml.funding_selection = "From job_config.yaml"
                    if job_config_yaml.environment_id:
                        job_config_yaml.environment_selection = "From job_config.yaml"
                    if job_config_yaml.image:
                        job_config_yaml.image_selection = "From job_config.yaml"

                    job_config = merge_config(job_config_yaml, job_config_decorator)
                else:
                    logger.debug("========no job_config.yaml found========")

                if job_config.commands:
                    console.print("[red]Error: The 'commands' configuration in job_config.yaml is not supported in the remote decorator.\nPlease remove it or run with 'robbie run' to use 'commands'.[/red]")
                    logger.error("The 'commands' configuration in job_config.yaml is not supported in the remote decorator.")
                    _exit_by_mode(1)
            
            job_config.job_type = JobRunType.REMOTE_FUNCTION_CALL
            logger.debug(job_config.to_string("job_config being passed to function"))
            
            console.print(f"Robbie is running your function remotely!", style="bold")
            return Deploy.remote_function_deploy(func, args, kwargs, job_config)

        return wrapper
    return decorator


import os
import time
import ipywidgets as widgets
from ipywidgets import Layout
from jupyter_ui_poll import ui_events
from IPython.display import display
from positron_common.cli.logging_config import logger
from positron_common.cli.console import console
from positron_common.job_api.funding_envs_images import *
from positron_common.config import PositronJob
from positron_common.cli_args import args

RUN_NAME_HELPER_TEXT = '(Optional customized run name)'
COMING_SOON_MENU_ITEM = 'Customize...(coming soon)'
clicked = False

def notebook_cell_ui() -> PositronJob:
    """ 
    When the user runs this function, it will display a UI that allows the user to select the hardware to run the notebook on.
    
    This function will display a UI that allows the user to select the hardware to run the notebook on.
    
    """
    #
    # Left Column
    #
    left_column_title = widgets.HTML(
        value="<b>Run your remote function on Robbie!</b>",
    )

    # logo
    file = open(os.path.join(os.path.dirname(__file__), 'img', 'dog.png'), "rb")
    image = file.read()

    logo = widgets.Image(
        value=image,
        format='png',
        width=100,
        height=100,
    )

    style = {'description_width': 'initial'}

    #
    # Middle Column
    #
    middle_column_title = widgets.HTML(
        value="<b>Just choose your hardware!</b>",
    )
     # Environment Drop down
    _drop_dict = _load_envs_for_dropdown()
    if _drop_dict is None:
        console.print("[red] Error loading environments")
        return
    _drop_dict.update({COMING_SOON_MENU_ITEM: None})
    
    hardware_dropdown = widgets.Dropdown(
        options = _drop_dict,
        # value=_drop_dict[1],
        description='Hardware Options:',
        style=style,
        layout={'width': 'max-content'}
    )

    _or_ = widgets.HTML(
        value="           - or - ",
    )

    """ robbie does it for you
    auto_checkbox = widgets.Checkbox(
        value=False,
        description='Let Robbie choose the best resources for you',
        disabled=False,
        indent=False
    )
    """

    # goes back to simple mode
    take_me_back_button = widgets.Button(
        description='Take me back',
        disabled=False,
        # button_style='info', # 'success', 'info', 'warning', 'danger' or ''
        tooltip='Go back to simple options',
        style = style
    )

    #
    # Right Column
    #
    right_column_title = widgets.HTML(
        value="<b>And get the ball rolling!</b>",
    )

    # optional name
    optional_name = widgets.Text(
        value=RUN_NAME_HELPER_TEXT,
        # description='Enter a name:',
        disabled=False,
        style=style
    )
    
    # "Run" button
    run_button = widgets.Button(
        description='Run',
        disabled=False,
        button_style='success', # 'success', 'info', 'warning', 'danger' or ''
        tooltip='Your notebook cell will run on Robbie',
        icon='play',
        style = style
    )

    coming_soon_text = widgets.HTML(
        value="Coming soon!",
    )

    # robbie does it for you
    tail_checkbox = widgets.Checkbox(
        value=False,
        description='Tail the remote machine stdout.',
        disabled=False,
        indent=False,
        style=style
    )

    # layout
    left_column = widgets.VBox([left_column_title, logo])
    middle_column = widgets.VBox([middle_column_title, hardware_dropdown, tail_checkbox],layout=Layout(margin = '0 20px 0 20px'))
    right_column = widgets.VBox([right_column_title, optional_name, run_button, ], layout=Layout(margin = '0 20px 0 20px'))
    ui = widgets.HBox([left_column, middle_column, right_column])

    out = widgets.Output()
    # display everything
    display(ui, out)

 
    def on_take_me_back_button_clicked(change):
        """ When we display the full list of option (all disabled) this button gets gets us back to the simple mode"""
        with out:
            # print("on_take_me_back_clicked() - change:", change)
            middle_column.children = [middle_column_title, hardware_dropdown, tail_checkbox]
    
    take_me_back_button.on_click(on_take_me_back_button_clicked)
    

    # this is the hardware dropdown to change to customization mode
    def on_hw_dropdown_changed(change):
        """ This is experimental now to show users what is coming"""
        with out:
            logger.debug(f"on_checkbox_clicked() - change:{change}")
            if change.get('type') == 'change' and change.get('name') == 'label' and change.get('new') == COMING_SOON_MENU_ITEM:
                logger.debug("detetected change")
                gpu_dropdown = widgets.Dropdown(
                    options=['Nvidia A10 (24GB)', 'Nvidia A100 (40GB)', 'Nvidia H100 (40GB)', 'CPU only'],
                    value='Nvidia A100 (40GB)',
                    description='GPU model:',
                    style=style,
                    layout={'width': 'max-content'},
                    disabled=True
                )
                cpu_slider=widgets.IntSlider(description='vCPU cores', min=1, max=64, step=1, value=4, style=style, disabled=True);
                memory_slider=widgets.IntSlider(description='System Memory (GB)', min=16, max=128, step=16, value=32, style=style, disabled=True);
                disksize_slider=widgets.IntSlider(description='Disk Size (GB)', min=16, max=128, step=16, value=32, style=style, disabled=True);
                middle_column.children = [middle_column_title, coming_soon_text, gpu_dropdown, cpu_slider, memory_slider, disksize_slider, take_me_back_button]
            else:
                logger.debug("did not detect change")

    hardware_dropdown.observe(on_hw_dropdown_changed)

    # this is the robbie does it for you checkbox
    """
    def on_checkbox_clicked(change):
        with out:
            # print("on_checkbox_clicked() - change:", change)
            if change.get('type') == 'change' and change.get('name') == 'value' and change.get('new') == True:
                hardware_dropdown.disabled = True
            elif change.get('type') == 'change' and change.get('name') == 'value' and change.get('new') == False:
                hardware_dropdown.disabled = False

    # register the event handler
    auto_checkbox.observe(on_checkbox_clicked)

    """
    
    def on_button_clicked(b):
        global clicked
        """ The user clicked the run button """
        print(f"on_button_clicked() - {clicked}")
        with out:
            logger.debug(f"Hardware dropdown value: {hardware_dropdown.value}")  
            if hardware_dropdown.value == None:
                console.print("[red] Please choose a hardware option, before running.")
                return
            else:
                # print(f"Running on {hardware_dropdown.value}")
                clicked = True
            

    # register the event handler
    run_button.on_click(on_button_clicked)

    # wait for the user to click the run button
    try:
        with ui_events() as poll:
            while not clicked:
                poll(10) # poll queued UI events including button
                time.sleep(1) # wait for 1 second before checking again
    except KeyboardInterrupt:
        console.print("[red] User interrupted.")
        return None

    # do we want to stream?
    if tail_checkbox.value:
        args.stream_stdout = True

    # return a job config object
    job_config = PositronJob()
    if optional_name.value != RUN_NAME_HELPER_TEXT:
        job_config.name=optional_name.value
    job_config.funding_group_id = None
    job_config.environment_id = hardware_dropdown.value
    job_config.environment_selection = "User selected from Dropdown"
    job_config.image = None 

    return job_config

#
# Reads the environments from the PERSONAL funding group
#

def _load_envs_for_dropdown():
    """
    Load the environments for the dropdown list in the cell output
    """
    fs = list_funding_sources()
    if len(fs) == 0:
        return None
    
    personal_fs_id = None
    for key, val in fs.items():
        if (val[FS_TYPE] == FS_PERSONAL_TYPE):
            personal_fs_id = val.get(FS_ID)
    
    if personal_fs_id is None:
        return None
    envs = list_environments(personal_fs_id)
    if len(envs) == 0:
        return None
    
    dropdown_list = {}
    for _, val in envs.items():
        dropdown_list.update({val.get(ENV_NAME) + " - " + val.get(ENV_DESCRIPTION) + f' ({val.get(ENV_TPH)} tokens/hr)' : val.get(ENV_ID)})
    return dropdown_list





    




import sys
import os
import subprocess
from positron_common.cli.console import console, print_boxed_messages
from positron_common.build_env import build_env
from positron_common.utils import get_version
from positron_common.env_defaults import current
from positron_common.env_config import env
from positron_common.user_config import user_config
from positron_common.cli_args import args

def common_dump():
    console.print("Dumping Robbie local environment...")
    print_boxed_messages(sys.executable, "Python executable (sys.executable)")
    print_boxed_messages(os.__file__, "Python executable (os.__file__)")
    print_boxed_messages(get_version(), "Robbie Python SDK version - importlib.metadata.version('robbie')")
    result = subprocess.run(["pip", "show", "robbie"], capture_output=True, text=True)
    print_boxed_messages(result.stdout, "Robbie SDK version (% pip show robbie)")
    result = subprocess.run(["pip", "list"], capture_output=True, text=True)
    print_boxed_messages(result.stdout, "Installed Python Packages (% pip list)")
    if os.path.exists("./requirements.txt"):
        try:
            with open("./requirements.txt", 'r') as file:
                # Read the content of the file
                file_content = file.read()
                print_boxed_messages(file_content, "./requirements.txt")
        except Exception as e:
            print(f"An error occurred: {e}")
        
    print_boxed_messages(str(build_env), "build_env")
    print_boxed_messages(env.to_string(), "EnvConfig (env)")
    print_boxed_messages(current.to_string(), "EnvDefaults (current)")
    print_boxed_messages(user_config.to_string(), "User Config (user_config)")     
    print_boxed_messages(args.to_string(), "CLI Arguments (args)")



'''
session_info finds and prints version information for loaded modules in the
current session, Python, and the OS.
'''

import sys
import types
import platform
import inspect
from datetime import datetime
from importlib import import_module
from inspect import cleandoc
from multiprocessing import cpu_count
from pathlib import Path

from stdlib_list import stdlib_list


def _imports(environ):
    '''Find modules in an environment.'''
    for name, val in environ:
        # If the module was directly imported
        if isinstance(val, types.ModuleType):
            yield val.__name__
        # If something was imported from the module
        else:
            try:
                yield val.__module__.split('.')[0]
            except AttributeError:
                pass


def _find_version(mod_version_attr):
    '''Find the version number of a module'''
    if (isinstance(mod_version_attr, str)
            or isinstance(mod_version_attr, int)):
        return mod_version_attr
    elif isinstance(mod_version_attr, tuple):
        joined_tuple = '.'.join([str(num) for num in mod_version_attr])
        return joined_tuple
    elif callable(mod_version_attr):
        try:
            return mod_version_attr()
        # A module might have a function that does something else than return
        # the version number altough it is named "version" or similar, e.g.
        # https://gitlab.com/joelostblom/session_info/-/issues/8. In these cases it
        # will likely fail to call without arguments, we're raising the same
        # type of error as in the parent try/except clause to continue looking
        # for more possible version strings.
        except Exception:
            raise AttributeError
    else:
        # print(f'Does not support module version of type {type(mod_ver_attr)}')
        return 'NA'


# From https://stackoverflow.com/a/52187331/2166823
# Slight modifications added
def _notebook_basename():
    """Returns the absolute path of the Notebook or None if it cannot be determined
    NOTE: works only when the security is token-based or there is also no password
    """
    import ipykernel
    import json
    import urllib
    from notebook import notebookapp

    connection_file = Path(ipykernel.get_connection_file()).name
    kernel_id = connection_file.split('-', 1)[1].split('.')[0]
    for srv in notebookapp.list_running_servers():
        if srv['token'] == '' and not srv['password']:  # No token and no password
            req = urllib.request.urlopen(srv['url']+'api/sessions')
        else:
            req = urllib.request.urlopen(srv['url']+'api/sessions?token='+srv['token'])
        sessions = json.load(req)
        for sess in sessions:
            if sess['kernel']['id'] == kernel_id:
                return Path(sess['notebook']['path']).stem
    return None


def _clean_modules(modules, private, excludes):
    # Keep module basename only. Filter duplicates and excluded modules.
    if private:
        clean_modules = set(module.split('.')[0] for module in modules
                            if module.split('.')[0] not in excludes)
    else:  # Also filter private modules
        clean_modules = set(module.split('.')[0] for module in modules
                            if module.split('.')[0] not in excludes
                            and not module.startswith('_'))
    return clean_modules


def show(na=True, os=True, cpu=False, jupyter=None, dependencies=None,
         std_lib=False, private=False, write_req_file=False, req_file_name=None,
         html=None, excludes=['builtins', 'stdlib_list'],caller_globals=None):
    '''
    Show version information for loaded modules in the current session,
    Python, and the OS.

    Parameters
    ----------
    na : bool
        Show module name even when no version number is found.
    os : bool
        Show OS information.
    cpu : bool
        Show number of logical CPU cores and info string (if available).
    jupyter : bool
        Show information about the jupyter environment. If `None`, output
        jupyter info only if inside a Jupyter notebook.
    dependencies : bool
        Show information about modules imported by the Python interpreter on
        startup and depency modules imported via other modules. If `None`,
        dependency modules will be included in the HTML output under a
        <details> tag, and excluded from the printed output. Setting `na` to
        `False` could be helpful to reduce verboseness.
    std_lib : bool
        Show information for modules imported from the standard library.
        Tries to detect the Python version to compare with the corresponding
        standard libarary, falls back to Python 3.7 if the version cannot be
        detected.
    private : bool
        Show information for private modules.
    write_req_file: bool
        Create a pip-compatible text file that lists all the module versions.
        If `None`, write dependency files for Jupyter notebooks only. If the
        `notebook` module is installed, `session_info` can automatically find the
        notebook name use it as a prefix for the requirments file. Otherwise
        it falls back to `session_info-requirements.txt`.
        This only writes explicitly imported modules.
    req_file_name : str
        Change the name of the requirements file.
    html: bool
        Format the output as HTML and collapse it in a <details> tag. If `None`,
        HTML will be used only if a Jupyter notebook environment is detected.
        Note that this will not be visible in notebooks shared on GitHub since
        they seemingly do not support the <details> tag. Requires IPython.
    excludes : list
        Do not show version information for these modules.
    '''
    # Exclude std lib packages
    if not std_lib:
        try:
            std_modules = stdlib_list(version=platform.python_version()[:-2])
        except ValueError:
            # Use 3.7 if the Python version cannot be found
            std_modules = stdlib_list('3.7')
        excludes = excludes + std_modules
    
    # Include jupyter info
    # Custom logic to detect if we're in a notebook, why didn't we just use `jupyter` option?
    in_notebook = 'jupyter_core' in sys.modules.keys()
    if in_notebook:
        if html is None:
            html = True
    if jupyter or (jupyter is None and in_notebook):
        jupyter = True
        jup_mod_names = ['IPython', 'jupyter_client', 'jupyter_core',
                         'jupyterlab', 'notebook']
        jup_modules = []
        for jup_mod_name in jup_mod_names:
            try:
                jup_modules.append(import_module(jup_mod_name))
            except ModuleNotFoundError:
                pass
        # The length of `'jupyter_client'` is 14
        # The spaces are added to create uniform whitespace in the output
        # f-strings, which is needed to clean them with inspect.cleandoc
        jup_mod_and_ver = [f'            {module.__name__:14}\t{module.__version__}'
                           for module in jup_modules]
        output_jupyter = '\n'.join(jup_mod_and_ver)
    else:
        output_jupyter = None

    # Custom logic to use a custom global object.
    if caller_globals is None:
        # Get `globals()` from the enviroment where the function is executed
        caller_globals = dict(
            inspect.getmembers(inspect.stack()[1][0]))["f_globals"]
    # Find imported modules in the global namespace
    imported_modules = _clean_modules(set(_imports(caller_globals.items())), private, excludes)
    # Wether to include dependency module not explicitly imported
    all_modules = {'imported': imported_modules}
    if dependencies is not False:  # Default with HTML is to include deps
        if html or dependencies:
            dependencies = True  # HTML default, used later for output strings
            sys_modules = _clean_modules(set(sys.modules.keys()), private, excludes)
            depend_modules = sys_modules.difference(imported_modules)
            if depend_modules:
                all_modules['depend'] = depend_modules
            else:
                dependencies = False

    output_modules = {}
    for mod_type in all_modules:
        clean_modules = all_modules[mod_type]

        # Don't duplicate jupyter module output
        if jupyter or in_notebook:
            for jup_module in jup_modules:
                if jup_module.__name__ in clean_modules:
                    clean_modules.remove(jup_module.__name__)

        # Find version number. Return NA if a version string can't be found.
        # This section is modified from the `py_session` package
        mod_and_ver = []
        mod_names = []
        mod_versions = []
        for mod_name in clean_modules:
            mod_names.append(mod_name)
            mod = sys.modules[mod_name]
            # Since modules use different attribute names to store version info,
            # try the most common ones.
            try:
                mod_version = _find_version(mod.__version__)
            except AttributeError:
                try:
                    mod_version = _find_version(mod.version)
                except AttributeError:
                    try:
                        mod_version = _find_version(mod.VERSION)
                    except AttributeError:
                        mod_version = 'NA'
                        # print(f'Cannot find a version attribute for {mod}.')
            mod_versions.append(mod_version)
        max_name_len = max([len(mod_name) for mod_name in mod_names] or [0])
        mod_and_ver = [f'{mod_name:{max_name_len}}\t{mod_version}'
                       for mod_name, mod_version in zip(mod_names, mod_versions)]
        if not na:
            mod_and_ver = [x for x in mod_and_ver if not x[-2:] == 'NA']
        mod_and_ver = sorted(mod_and_ver)
        output_modules[mod_type] = '\n            '.join(mod_and_ver)

    # Write requirements file for notebooks only by default
    if write_req_file or (write_req_file is None and in_notebook):
        if req_file_name is None:
            # Only import notebook libraries if we are in the notebook
            # Otherwise, running this multiple times would set `in_notebook=True`
            if in_notebook:
                try:
                    req_file_name = f'{_notebook_basename()}-requirements.txt'
                # If the notebook name cannot be found for, we want to fall
                # back to the default name. Usually this happens because the
                # notebook module is not found, but this is already mentioned
                # in the docstring so we're not differentiating between that
                # and other exceptions here.
                except Exception:
                    req_file_name = 'session_info-requirements.txt'
            else:
                req_file_name = 'session_info-requirements.txt'

        # For NA modules, just include the latest, so no version number.
        mods_na_removed = [mod_ver.replace('\tNA', '')
                           for mod_ver in output_modules['imported'].split('\n')]
        if jupyter:
            mods_req_file = mods_na_removed + jup_mod_and_ver
        else:
            mods_req_file = mods_na_removed
        clean_mods_req_file = [mod_ver.replace(' ', '').replace('\t', '==')
                               for mod_ver in mods_req_file]
        with open(req_file_name, 'w') as f:
            for mod_to_req in clean_mods_req_file:
                f.write('{}\n'.format(mod_to_req))

    # Sys info
    sys_output = 'Python ' + sys.version.replace('\n', '')
    os_output = platform.platform() if os else ''
    if cpu:
        if platform.processor() != '':
            cpu_output = f'{cpu_count()} logical CPU cores, {platform.processor()}'
        else:
            cpu_output = f'{cpu_count()} logical CPU cores'
    else:
        cpu_output = ''
    date_output = 'Session information updated at {}'.format(
        datetime.now().strftime('%Y-%m-%d %H:%M'))

    # Output
    # For proper formatting and exclusion of `-----` when `jupyter=False`.
    nl = '\n'
    output_jup_str = ('' if output_jupyter is None
                      else f'{nl}            -----{nl}{output_jupyter}')
    if html:
        from IPython.display import HTML
        if dependencies:
            # Must be dedented to line up with the returned HTML.
            # Otherwise `cleandoc()` does not work.
            output_depend_str = f"""
            -----
            </pre>
            <details>
            <summary>Click to view modules imported as dependencies</summary>
            <pre>
            {output_modules['depend']}
            </pre>
            </details> <!-- seems like this ends pre, so might as well be explicit -->
            <pre>"""
        else:
            output_depend_str = ''
        return HTML(cleandoc(f"""
            <details>
            <summary>Click to view session information</summary>
            <pre>
            -----
            {output_modules['imported']}{output_depend_str}{output_jup_str}
            -----
            {sys_output}
            {os_output}
            {cpu_output}
            -----
            {date_output}
            </pre>
            </details>""").replace('\n\n', '\n'))  # Clean empty output fields
    else:
        if dependencies:
            # Must be dedented to line up with the returned HTML.
            # Otherwise `cleandoc()` does not work.
            output_depend_str = f"""
            -----
            {output_modules['depend']}"""
        else:
            output_depend_str = ''
        print(cleandoc(f"""
            Session information:
            -----
            {output_modules['imported']}{output_depend_str}{output_jup_str}
            -----
            {sys_output}
            {os_output}
            {cpu_output}
            -----
            {date_output}""").replace('\n\n', '\n'))  # Clean empty output fields



import typer
from positron_cli.run import run
from positron_cli.login import login
from positron_cli.set_env import set_env
from positron_cli.configure import configure
from positron_cli.download import download
from positron_cli.set_log_level import log_level
from positron_cli.version import version
from positron_cli.cmd_dump import dump
from positron_common.cli.console import console
import pyfiglet

ascii_banner = pyfiglet.figlet_format("Robbie")
console.print(ascii_banner, style='#41a7ff')
app = typer.Typer(help="A CLI tool to help you run your code in the Robbie")

app.command()(dump)
app.command()(run)
app.command()(login)
app.command()(set_env)
app.command()(configure)
app.command()(download)
app.command()(version)
app.callback()(log_level)

if __name__ == "__main__":
    app()



import time
import signal
import os
from rich.prompt import Confirm
from positron_common.exceptions import RemoteCallException
from positron_common.env_config import env
from positron_common.config import PositronJob
from positron_common.cli_args import args as cli_args
from positron_common.print import *
from positron_common.cli.console import console, ROBBIE_BLUE
from positron_common.constants import temp_path
from positron_common.compression import create_workspace_tar
from positron_common.cli.logging_config import logger
from positron_common.job_api.stream_logs import start_stdout_stream
from positron_common.job_api.get_job import get_job
from positron_common.job_api.start_job import start_job
from positron_common.job_api.terminate_job import terminate_job
from positron_common.job_api.create_job import create_job
from positron_common.utils import get_default_workspace_dir, _exit_by_mode, SUCCESS, FAILURE
from positron_common.aws.s3_presigned_handler import S3PresignedHandler
from positron_cli.download import download_chosen_results

# number of seconds to poll when monitoring job status
POLLING_SEC=1

# for the the deep link
PORTAL_BASE = env.API_BASE.rstrip('/api')
    
def command_runner_deploy(job_config: PositronJob):
    """
    This function is the entry point for the running a command on a remote machine.
    It can be called in two ways:
    - Mode 1 - From the command line with the `positron run` command
    - Mode 2 - From the run_notebook() function when running in a notebook (Jupyter, Colab, VScode
    """
    signal.signal(signal.SIGINT, handle_sigint)
    # logger.debug(env)
    
    try:
        logger.debug(job_config.to_string("command_runner_deploy"))
        job_config.validate_values()

        # TODO: We should not be creating a job before we let the user run it, we need defaults in the DBs that we can query
        logger.debug(job_config.create_runtime_env())
        job = create_job(job_config=job_config)
        
        # print the configuration banner
        print_robbie_configuration_banner(job, job_config)
    
        # prompt the user if they don't pass the -y option
        if not cli_args.skip_prompts:
            user_input = input("Run job with these settings? ([y]/n)")
            if not user_input in ["yes", "y", "Yes", "Y", ""]:
                terminate_job(job["id"], "User declined from CLI")
                console.print("[yellow]See you soon![/yellow]")
                return

        # tell people we are on the local machine
        console.print("[bold]Local Machine: [/bold]", style=ROBBIE_BLUE)    

        workspace_dir = (job_config.workspace_dir if job_config.workspace_dir else get_default_workspace_dir())
        logger.debug(f'Workspace directory: {workspace_dir}')
        
        if os.path.exists(workspace_dir):
            console.print(f'Workspace directory: {workspace_dir}')
        else:
            console.print(f"[bold red] ERROR: Workspace directory does not exist: {workspace_dir}")
            return

        # compress the workspace
        file_count = create_workspace_tar(workspace_dir=workspace_dir)
        console.print("[green]✔[/green] Packaged up workspace artifacts...(1 of 3)", style=ROBBIE_BLUE)

        if file_count == 0:
            Confirm.ask("No files were found in the workspace directory. Would you like to continue anyway?", default=False)
    
        # upload the compressed workspace
        S3PresignedHandler.upload_file_to_job_folder(f"{temp_path}/{env.COMPRESSED_WS_NAME}", job['id'], env.COMPRESSED_WS_NAME)
        console.print("[green]✔[/green] Uploaded compressed workspace to Robbie...(2 of 3)", style=ROBBIE_BLUE)
        
        if cli_args.create_only:
            console.print(f"[green]✔[/green] Job created successfully. (3 of 3)")
            console.print(f"JOB_ID: {job.get('id')}")
            return

        # start the job up
        start_job(job_id=job['id'], data=job_config.create_runtime_env())
        console.print("[green]✔[/green] Submitted job to Robbie...(3 of 3)", style=ROBBIE_BLUE)

        start = time.perf_counter()
        print_job_details_banner(job)

        # Are we streaming stdout or just showing the status changes.
        if cli_args.stream_stdout:
            # tell people we are on the remote machine
            console.print("[bold]Remote Machine Status: [/bold]", style=ROBBIE_BLUE)  
            start_stdout_stream(job['id'])

            final_get_job = get_job(job['id'])
            # did someone interrup the stream?
            if _is_job_done(final_get_job):
                print_job_complete_banner(final_get_job, start)
                if cli_args.download:
                    # download the results
                    download_chosen_results(final_get_job['id'], cli_args.download, cli_args.local_path)
            
                if _was_job_a_success(final_get_job):
                    _exit_by_mode(SUCCESS)
                else:
                    _exit_by_mode(FAILURE)
        else:
            if cli_args.monitor_status:
                # lets track and display the status updates
                console.print(f"You can also monitor job status in the Robbie portal at: {PORTAL_BASE}/portal/app/my-runs?jobId={job['id']}\n", style=ROBBIE_BLUE) 
                
                # tell people we are on the remote machine
                console.print("[bold]Remote Machine Status: [/bold]", style=ROBBIE_BLUE)  
                last_status_change = "Starting..."
                final_get_job = None
        
                console.print("Processing...", style=ROBBIE_BLUE)   
                while True:
                    job_result = get_job(job['id'])
                    # are we in a final state?
                    if(_is_job_done(job_result)):
                        break
                    if(job_result['status'] != last_status_change):
                        # there has been a status change
                        time1 = time.strftime("%H:%M:%S")
                        console.print(f"\t{time1}: {job_result['status']}")
                        last_status_change = job_result['status']
                    time.sleep(POLLING_SEC)
                # job is down now, diplay final results.    
                final_get_job = get_job(job['id'])
                print_job_complete_banner(final_get_job, start)
                if cli_args.download:
                    # download the results
                    download_chosen_results(final_get_job['id'], cli_args.download, cli_args.local_path)
                
                if _was_job_a_success(final_get_job):
                    _exit_by_mode(SUCCESS)
                else:
                    _exit_by_mode(FAILURE)
            else:
                console.print(f"You can monitor job status in the Robbie portal at: {PORTAL_BASE}/portal/app/my-runs?jobId={job['id']}", style=ROBBIE_BLUE) 
                _exit_by_mode(SUCCESS)
    except RemoteCallException as e:
        """For known errors we dont print exceptions, we just print the user friendly message"""
        print_known_error(e)
        _exit_by_mode(FAILURE)
    except Exception as e:
        # don't let this propogate up, we want to catch all exceptions
        logger.exception(e)
        print(e)
        _exit_by_mode(FAILURE)

def _is_job_done(job) -> bool:
    return job['status'] == "terminated" or job['status'] == "complete" or job['status'] == "failed" or job['status'] == "execution_error"

def _was_job_a_success(job) -> bool:
    return job['status'] == "complete"

def handle_sigint(signum, frame):
    console.print('Terminating gracefully...')
    # TODO: we should actually close open connections
    _exit_by_mode(FAILURE)








import asyncio
import socketio
import nest_asyncio
import json
import time
from positron_common.env_config import env
from positron_common.exceptions import RobbieException
from ..cli.console import console, ROBBIE_BLUE
from ..cli.logging_config import logger

sio = socketio.AsyncClient()
nest_asyncio.apply() # enabled nested event loops

@sio.event(namespace='/stdout-stream')
async def connect():
    console.print("Connected to your run's log stream!")

@sio.event(namespace='/stdout-stream')
async def message(message: str):
    try:
        log: dict = json.loads(message)
        level_name = log.get('log_level')
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S:%MS", time.gmtime(log.get('timestamp') / 1000))
        message = log.get('message')
        level_string = level_name.ljust(8)
        logger_name = log.get('app_name')

        if level_name == "INFO":
            level_string = f"[green]{level_name}[/green]"
        elif level_name == "DEBUG":
            level_string = f"[blue]{level_name}[/blue]"
        elif level_name == "ERROR":
            level_string = f"[red]{level_name}[/red]"
        elif level_name == "WARNING":
            level_string = f"[yellow]{level_name}[/yellow]"

        formatted_log = f"{logger_name}: {timestamp} {level_string}: {message}"
        console.print(formatted_log)
    except:
        console.print(log['message'], style=ROBBIE_BLUE)

@sio.event(namespace='/stdout-stream')
async def disconnect():
    console.print('Disconnected from stdout stream')

@sio.event(namespace='/stdout-stream')
async def error(err):
    console.print('An error occurred in the streaming process')
    logger.error(err)

async def start_stream(job_id: str):
    custom_headers = {
        "PositronAuthToken": env.USER_AUTH_TOKEN,
        "PositronJobId": job_id
    }
    await sio.connect(env.SOCKET_IO_DOMAIN, headers=custom_headers, socketio_path=env.SOCKET_IO_PATH)
    # TODO: I don't think we want to actually do this long term.
    try:
        await sio.wait()
    except asyncio.exceptions.CancelledError as error:
        pass # user will be notified that the stream has ended in the disconnect event
    

def start_stdout_stream(job_id: str):
    try:
        # Start the stream
        asyncio.get_event_loop().run_until_complete(start_stream(job_id))
    except Exception as error:
        raise RobbieException(error)





import requests
import os
import tarfile
import typer
import uuid
from typing_extensions import Annotated
from typing import Optional
from tqdm.auto import tqdm
from rich.tree import Tree
from positron_common.job_api.list_jobs import list_jobs
from positron_common.job_api.get_job import get_job
from positron_common.aws.s3_presigned_handler import S3PresignedHandler
from positron_common.exceptions import RobbieException, RemoteCallException
from positron_common.cli.console import console, ROBBIE_BLUE
from positron_common.cli.logging_config import logger
from positron_common.observability.main import track_command_usage
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.validation import Validator

COMPRESSED_RESULTS = 'result.zip'

@track_command_usage("download")
def download(
    job_id_arg: Annotated[Optional[str], typer.Option("--run", help='Run UUID')] = None,
    filename: Annotated[Optional[str], typer.Option("--filename", help='Name of the file to download or all for a tar file of all results.')] = None,
    local_path: Annotated[Optional[str], typer.Option("--path", help='Path of where to download the file.')] = ".",
) -> None:
    """
        Download command for Robbie CLI ("robbie download) allows users to download result.zip from a job.
        
        Prompts the user to choose a job to download results from.
    """
    try:
        if not filename_is_valid(filename):
            console.print('[red]Error: Please specify a valid file name or "all" to download all files.')
            return
        
        if local_path and not os.path.exists(local_path):
            console.print('[red]Error: The path you specified is not valid.')
            return

        if job_id_arg:
            if is_valid_uuid(job_id_arg):
                download_chosen_results(job_id_arg, filename, local_path)
            else:
                console.print("[red]Error: Please enter a valid run id in UUID format.")
        else:
            # prompt for the job to download
            jobs = Jobs.load()

            def is_job_valid(text):
                return text == "" or text in jobs.menu_items()
            job_validator = Validator.from_callable(is_job_valid, error_message='Please enter a valid run name.')

            jobs_choice = prompt('Choose a run to download: <tab for menu>: ', completer=WordCompleter(jobs.menu_items()), validator=job_validator)
            if jobs_choice == "" or jobs_choice == None:
                console.print("No job selected. Exiting.")
                return
            download_chosen_results(jobs.id_from_menu_item(jobs_choice), filename, local_path)
    except RobbieException as e:
        console.print(f"[red]Error: {e}[/red]")
    except RemoteCallException as e:
        console.print(f"[red]Error: {e}[/red]")
    
def filename_is_valid(filename: str) -> bool:
    """Check if the download argument is valid."""
    if filename == None:
        return False
    if filename == "all":
        return True
    if filename == "":
        return False
    return True

def download_chosen_results(job_id: str, filename: str, local_path: str):
    """The user can download a single file or all files from the run."""
    if filename == "all":
        download_tar_results_file(job_id, local_path)
    elif filename == None or filename == "":
        raise RobbieException(f"Nothing to download run {job_id}")
    else:
        download_individual_results_file(job_id, filename, local_path)

def download_tar_results_file(job_id: str, local_path: str = "."):
    """
    Download the result.zip file from the job and stores it in the CWD

    Raises:
        RobbieException: If we can't generate a presigned URL or downloading fails
    """
    job = get_job(job_id)
    if(job == None):
        raise RobbieException(f"No such run: {job_id}")

    if(job["resultsPresignedBucketInfo"] == None):
        raise RobbieException(f"No resultsPresignedBucketInfo for job {job_id}")

    logger.debug("resultsPresignedBucketInfo", job["resultsPresignedBucketInfo"])

    logger.debug(f'Downloading results for: {job["name"]}')

    response = requests.get(job["resultsPresignedBucketInfo"],stream=True) 
    if response.status_code != 200:
        logger.debug(f'Failed to download URL, http code: {response.status_code} \n {response.text}')
        raise RobbieException('Sorry, run has no results to download.') 
    else:
        # Sizes in bytes.
        total_size = int(response.headers.get("content-length", 0))
        block_size = 1024

        console.print(f'Download results for: {job["name"]}')

        open_path = f"{local_path}/{COMPRESSED_RESULTS}"
        console.print(f"Downloading file: {COMPRESSED_RESULTS} for run {job_id}, to {open_path}")
        with tqdm(total=total_size, unit="B", unit_scale=True) as progress_bar:
            with open(open_path, "wb") as file:
                logger.debug(f'Successfull opened: {open_path}')
                for data in response.iter_content(block_size):
                    progress_bar.update(len(data))
                    file.write(data)

        if total_size != 0 and progress_bar.n != total_size:
            raise RobbieException(f'failed to download file')
    
        console.print("[green]✔[/green] Results now available.")


def download_individual_results_file(job_id: str, file_name: str, local_path: str = "."):
    """
    Download a single file from the job and stores it in the CWD

    Raises:
        RobbieException: If we can't generate a presigned URL or downloading fails
    """
    
    file_url = S3PresignedHandler._get_download_presigned_url(job_id, f"result/{file_name}")
    if(file_url == None):
        raise RobbieException(f"No such file: {file_name} for run {job_id}")

    logger.debug("file_url", file_url)
    logger.debug(f"Downloading file: {file_name} for run {job_id}")

    # _get_download_presigned_url returns a dict with a 'url' key
    response = requests.get(file_url['url'], stream=True) 
    if response.status_code != 200:
        logger.debug(f'Failed to download URL, http code: {response.status_code} \n {response.text}')
        raise RobbieException(f'Sorry, {file_name} does not exist for run: {job_id}.') 
    else:
        # Sizes in bytes.
        total_size = int(response.headers.get("content-length", 0))
        block_size = 1024

        open_path = f"{local_path}/{file_name}"
        console.print(f"Downloading file: {file_name} for run {job_id}, to {open_path}")
        with tqdm(total=total_size, unit="B", unit_scale=True) as progress_bar:
            with open(open_path, "wb") as file:
                logger.debug(f'Successfull opened ./{open_path}')
                for data in response.iter_content(block_size):
                    progress_bar.update(len(data))
                    file.write(data)

        if total_size != 0 and progress_bar.n != total_size:
            raise RobbieException(f'failed to download file: {open_path}')
        
        '''
        # perserving this code if we decide to change the functionality back to auto decompress .zip files

        _unzip(zip_file="./result.zip", path=".")
        os.remove("./result.zip")
        '''
    
        console.print("[green]✔[/green] File downloaded successfully.")

'''
# perserving this function is we decide to change the functionality back to auto decompress .zip files
def _unzip(zip_file, path):
    """
    Extracts `zip_file` and puts the `members` to `path`.
    If members is None, all members on `zip_file` will be extracted.
    """
    tree = Tree("Remote files copied to Local Machine")

    with zipfile.ZipFile(zip_file, "r") as zip_ref:
        for member in zip_ref.namelist():
            zip_ref.extract(member, path=path)
            tree.add(f"[yellow]{member}, size: {os.path.getsize(member)} bytes[/yellow]")
    console.print(tree)
'''

def is_valid_uuid(uuid_to_test, version=4):
    try:
        # check for validity of Uuid
        uuid_obj = uuid.UUID(uuid_to_test, version=version)
    except ValueError:
        return False
    return True

# Naming
JOB_ID="id"
JOB_NAME="name"
JOB_MENU="menu"

# singleton builds a list of tuples from the DB results
class Jobs: 
    is_init: bool = False
    my_jobs: dict

    def __init__(self, jobs_arg: dict):
        if self.is_init:
            raise ValueError('Jobs.load() already initialized')
        else:
            self.init = True
            self.my_jobs= jobs_arg

    @staticmethod
    def load():
        jobs = list_jobs()
        if len(jobs) == 0:
            return None
        # Loop through and add a custom "menu" item to each dict (but only if the job actually ran)
        for key, val in jobs.items(): 
            if val["durationMs"] != None:
                val[JOB_MENU] = f'{val[JOB_NAME]} (uuid: {val[JOB_ID]} )'
        return Jobs(jobs)
        
    # Prompt toolkit needs a list of strings to display in the menu 
    def menu_items(self) -> list: 
        ret_list: list = []
        for key, val in self.my_jobs.items():
            # just show names
            if val["durationMs"] != None:
                ret_list.append(val[JOB_MENU])
        return ret_list
    
    def id_from_menu_item(self, menu_item: str) -> str:
        for key, val in self.my_jobs.items():
            if val["durationMs"] != None and val[JOB_MENU] == menu_item:
                return val[JOB_ID]
        return None


from positron_common.utils import _current_python_version
from positron_common.constants import *
from positron_common.env_defaults import current
from positron_common.cli.logging_config import logger
from positron_common.job_api.funding_envs_images import *

# used to construct the image name
IMAGE_VERSION = '1.0.0'
BASE_NAME = "robbie"
LINUX_VERSION = "ubuntu22.04"

DEFAULT_PYTORCH_VERSION_GPU = {
    "py3.9": "2.0",
    "py3.10": "2.2",
    "py3.11": "2.4",
    "py3.12": "2.5",
}

DEFAULT_PYTORCH_VERSION_CPU = {
    "py3.9": "1.13",
    "py3.10": "2.2",
    "py3.11": "2.4",
    "py3.12": "2.5",
}

def get_auto_image_name_and_cluster(fs_id: str, env_id: str):
    """
    Constructs an image name based on the current environment, chosen cluster, and python version
    """
    logger.debug(f"get_auto_image_name_and_cluster(): {fs_id}, {env_id}")

    gpu_count, cluster_type = get_gpu_count_and_cluster_from_env(fs_id, env_id)
    # if there are GPUs in the environment, use the GPU image
    if gpu_count:
        type = "gpu"
    else:
        type = "cpu"
    logger.debug(f"type: {type}")

    # if running local use dev images
    if current.name == "localhost":
        env = "dev"
    else:
        env = current.name

    python_version = f"py{_current_python_version()}"
    logger.debug(f"Python version string: {python_version}")
    if type == "gpu" and not DEFAULT_PYTORCH_VERSION_GPU.get(python_version):
        logger.debug(f"No default PyTorch GPU version found for: {python_version}")
        return None, None
    if (type == "cpu" and not DEFAULT_PYTORCH_VERSION_CPU.get(python_version)):
        logger.debug(f"No default PyTorch CPU version found for: {python_version}")
        return None, None
    
    torch_version = DEFAULT_PYTORCH_VERSION_GPU.get(python_version) if type == "gpu" else DEFAULT_PYTORCH_VERSION_CPU.get(python_version)
    image_name = f"{BASE_NAME}:{IMAGE_VERSION}-{type}-{python_version}-torch{torch_version}-{LINUX_VERSION}-{env}"
    logger.debug(f"Automatically generated image name: {image_name}")
    return image_name, cluster_type

    
   
def get_gpu_count_and_cluster_from_env(fs_id: str, env_id: str):
    """ Gets the cluster type from the environment """
    envs = list_environments(fs_id)
    if len(envs) == 0:
        return None, None
    for _, val in envs.items():
        if(val.get(ENV_ID) == env_id):
            logger.debug(f"Environment: {val.get(ENV_NAME)}, Id: {val.get(ENV_ID)}, Cluster type: {val.get(ENV_CLUSTER_TYPE)}, GPU count: {val.get(ENV_GPU_NUMBER)}")
            return val.get(ENV_GPU_NUMBER), val.get(ENV_CLUSTER_TYPE)
    return None, None

      


import io
from typing import List
import matplotlib.image as mpimg
import matplotlib.pyplot as plt
from positron_common.job_api.get_job import get_job
from positron_common.aws.s3_presigned_handler import S3PresignedHandler

class ImageResult:
    def __init__(self, job_id: str):
        self.image_paths = []
        self.job_id = job_id

    def has_images(self) -> bool:
        job = get_job(self.job_id)

        result_image_file_names = job.get("resultImageFileNames")

        if result_image_file_names:
            self.image_paths = result_image_file_names
            return len(self.image_paths) > 0

        return False

    def _load_image_paths(self) -> List[bytes]:
        if not self.image_paths:
            raise ValueError("No image paths found")
        
        images = [] 
        for image_path in self.image_paths:
            image_bytes = S3PresignedHandler.download_file_to_bytes(self.job_id, image_path)
            images.append(image_bytes)
        
        return images

    def show_images(self):
        images = self._load_image_paths()
        for image in images:
            img = mpimg.imread(io.BytesIO(image), format='png')
            plt.imshow(img)
            plt.axis('off')
            plt.show()
